# @Created Date: 2020-09-28 05:42:53 pm
# @Filename: record.py
# @Email:  1730416009@stu.suda.edu.cn
# @Author: ZeFeng Zhu
# @Last Modified: 2020-09-28 05:43:34 pm
# @Copyright (c) 2020 MinghuiGroup, Soochow University
from pdb_profiling.processors.ensembl.api import EnsemblAPI
from pdb_profiling.processors.eutils.api import EutilsAPI
from pdb_profiling.processors.proteins.api import ProteinsAPI
from pdb_profiling.processors.proteins import ProteinsDB
from pdb_profiling.log import Abclog
from pdb_profiling.utils import init_folder_from_suffix, a_seq_reader, a_load_json, init_semaphore, unsync_wrap, unsync_run
from re import compile as re_compile
from pathlib import Path
from typing import Union, Optional, Tuple, Iterable, Callable, List
from unsync import unsync
from asyncio import as_completed
from pandas import DataFrame, concat
from collections import OrderedDict


class Identifier(Abclog):
    suffix = r'[0-9]+)[\.]*([0-9]*)'
    pats = OrderedDict({
        ('RefSeq', 'model'): re_compile('(X[A-Z]{1}_{%s}' % suffix),
        ('RefSeq', 'transcript'): re_compile(f'(NM_{suffix}'),
        ('RefSeq', 'protein'): re_compile(f'(NP_{suffix}'),
        ('Ensembl', 'gene'): re_compile(f'(ENSG{suffix}'),
        ('Ensembl', 'transcript'): re_compile(f'(ENST{suffix}'),
        ('Ensembl', 'protein'): re_compile(f'(ENSP{suffix}'),
        ('UniProt', 'isoform'): re_compile(r'^((?=.*[A-Za-z])(?=.*\d)[A-Za-z\d]{6,})[\-]*([0-9]*)$')
        })

    @classmethod
    @unsync
    async def set_web_semaphore(cls, *web_semaphore_values):
        if len(web_semaphore_values) == 0:
            proteins_api, ensembl_api, eutils_api = 20, 20, 20
        elif len(web_semaphore_values) < 3:
            num = web_semaphore_values[0]
            proteins_api, ensembl_api, eutils_api = num, num, num
        else:
            proteins_api, ensembl_api, eutils_api = web_semaphore_values[:3]
        cls.proteins_api_web_semaphore = await init_semaphore(proteins_api)
        cls.ensembl_api_web_semaphore = await init_semaphore(ensembl_api)
        cls.eutils_api_web_semaphore = await init_semaphore(eutils_api)

    @classmethod
    def set_folder(cls, folder: Union[Path, str]):
        cls.folder = Path(folder)
        cls.sqlite_api = ProteinsDB(
            "sqlite:///%s" % (init_folder_from_suffix(cls.folder, 'local_db')/"proteinsAPI.db"))
        cls.proteins_api_folder = init_folder_from_suffix(
            cls.folder, 'proteins/api/proteins')
        cls.seq_folder = dict(
            RefSeq=init_folder_from_suffix(cls.folder, 'eutils/efetch'),
            Ensembl=init_folder_from_suffix(cls.folder, 'ensembl/sequence/id'))
        cls.ensembl_archive_folder = init_folder_from_suffix(
            cls.folder, 'ensembl/archive/id')

    @classmethod
    def get_type(cls, identifier: str):
        for key, pat in cls.pats.items():
            res = pat.fullmatch(identifier)
            if bool(res):
                return key, res.groups()

    def __init__(self, identifier: str, folder: Optional[Union[Path, str]] = None):
        try:
            (self.source, self.level), (self.identifier,
                                        self.version) = self.get_type(identifier)
            self.raw_identifier = identifier
            if folder is not None:
                self.set_folder(folder)
            getattr(self, 'sqlite_api')
        except TypeError:
            raise ValueError(f"Unexpected identifier type: {identifier}")
        except AttributeError:
            raise AttributeError(
                "Please specify class variable `folder` via set_folder() first or pass `folder` in this method!")
        self.status = None

    def __repr__(self):
        return f'<{self.source} {self.level} {self.identifier} {self.version}>'

    @unsync
    async def set_status(self):
        if self.source == 'Ensembl':
            res = await EnsemblAPI.single_retrieve('archive/id/',
                self.identifier,
                None,
                self.ensembl_archive_folder,
                self.ensembl_api_web_semaphore,
                headers={'Content-Type': 'application/json'})
            if res is None:
                self.status = False
            else:
                self.status = await a_load_json(res)

    @classmethod
    def save_ProteinsAPI_data_to_DB(cls, res, identifier=''):
        dbReferences_df, other_dbReferences_df, iso_df, features_df, int_df, info_data = res

        cls.sqlite_api.sync_insert(
            cls.sqlite_api.INFO, [info_data])

        cls.sqlite_api.sync_insert(
            cls.sqlite_api.FEATURES, features_df.to_dict('records'))

        if (dbReferences_df is not None) and (len(dbReferences_df) > 0):
            cls.sqlite_api.sync_insert(
                cls.sqlite_api.DB_REFERENCES, dbReferences_df.to_dict('records'))
            if iso_df is not None:
                cls.sqlite_api.sync_insert(
                    cls.sqlite_api.ALTERNATIVE_PRODUCTS, iso_df.to_dict('records'))
            else:
                cls.logger.info(
                    f"Can't find ALTERNATIVE_PRODUCTS with {identifier}")
        else:
            cls.logger.info(
                f"Can't find (reviewed) dbReference with {identifier}")

        if len(other_dbReferences_df) > 0:
            cls.sqlite_api.sync_insert(
                cls.sqlite_api.OTHER_DB_REFERENCES, other_dbReferences_df.to_dict('records'))
        if int_df is not None and len(int_df) > 0:
            cls.sqlite_api.sync_insert(
                cls.sqlite_api.INTERACTION, int_df.to_dict('records'))

    @unsync
    async def fetch_from_ProteinsAPI_with_unp(self):
        assert self.source == 'UniProt'
        res = await ProteinsAPI.pipe_summary([await ProteinsAPI.single_retrieve(
            'proteins/',
            dict(),
            self.proteins_api_folder,
            self.proteins_api_web_semaphore,
            identifier=self.identifier
        ).then(a_load_json)])
        if res is None:
            return
        else:
            self.save_ProteinsAPI_data_to_DB(res, identifier=self.identifier)
            return res

    @unsync
    async def query_from_DB_with_unp(self, table_name:str, columns: str = '*', exists:Optional[bool]=None):
        default_tables = ('DB_REFERENCES', 'OTHER_DB_REFERENCES', 'ALTERNATIVE_PRODUCTS', 'FEATURES', 'INTERACTION', 'INFO')
        assert table_name in default_tables
        exists = (await self.sqlite_api.INFO.objects.filter(accession=self.identifier).exists()) if exists is None else exists
        if not exists:
            res = await self.fetch_from_ProteinsAPI_with_unp()
            res = dict(zip(default_tables, res))
            return res[table_name]
        else:
            if columns == '*':
                return DataFrame(
                    await getattr(self.sqlite_api, table_name).objects.filter(accession=self.identifier).all())
            else:
                return DataFrame(
                    (await self.sqlite_api.database.fetch_all(query=f"SELECT {columns} FROM {table_name} WHERE accession = '{self.identifier}'")),
                    columns=columns.split(','))
            '''
            dbReferences_df = DataFrame(await cls.sqlite_api.DB_REFERENCES.objects.filter(accession=accession).all())
            other_dbReferences_df = DataFrame(await cls.sqlite_api.OTHER_DB_REFERENCES.objects.filter(accession=accession).all())
            iso_df = DataFrame(await cls.sqlite_api.ALTERNATIVE_PRODUCTS.objects.filter(accession=accession).all())
            features_df = DataFrame(res)
            int_df = DataFrame(await cls.sqlite_api.INTERACTION.objects.filter(accession1__contains=accession).all())
            return dbReferences_df, other_dbReferences_df, iso_df, features_df, int_df
            '''

    @unsync
    async def fetch_from_ProteinsAPI(self, reviewed='true', isoform=0):
        res = await ProteinsAPI.pipe_summary(await ProteinsAPI.single_retrieve(
            'proteins/',
            dict(offset=0, size=-1, reviewed=reviewed, isoform=isoform),
            self.proteins_api_folder,
            self.proteins_api_web_semaphore,
            identifier=f'{self.source}:{self.identifier}'
        ).then(a_load_json))
        if res is None:
            return
        else:
            self.save_ProteinsAPI_data_to_DB(res, identifier=self.identifier)
            return res

    @unsync
    async def get_all_level_identifiers(self):
        try:
            cur_id = self.raw_identifier if self.source == 'RefSeq' else self.identifier
            return dict(zip(('protein', 'transcript', 'gene'), await self.sqlite_api.database.fetch_one(
                query=f"""
                    SELECT protein,transcript,gene FROM dbReferences
                    WHERE type == '{self.source}' AND ({self.level} == '{cur_id}' OR {self.level} LIKE '{cur_id}%')""")))
        except TypeError:
            return

    @unsync
    async def map2unp_from_DB(self):
        '''
        return accession, UniProt Isoform, is_canonical
        '''
        cur_id = self.raw_identifier if self.source == 'RefSeq' else self.identifier
        res = await self.sqlite_api.database.fetch_one(
            query=f"""
                SELECT accession,isoform FROM dbReferences
                WHERE type == '{self.source}' AND ({self.level} == '{cur_id}' OR {self.level} LIKE '{cur_id}%')""")
        if res is None:
            return
        else:
            accession, isoform = res
        if isoform is not None:
            sequenceStatus = await self.sqlite_api.database.fetch_val(
                query=f"""
                    SELECT sequenceStatus FROM ALTERNATIVE_PRODUCTS 
                    WHERE accession == '{accession}' AND isoform == '{isoform}'""")
            assert sequenceStatus is not None, accession
            return self.raw_identifier, accession, isoform, sequenceStatus == 'displayed'
        else:
            sequenceStatus = await self.sqlite_api.database.fetch_one(
                query=f"""
                    SELECT sequenceStatus FROM ALTERNATIVE_PRODUCTS 
                    WHERE accession == '{accession}'""")
            if sequenceStatus is None:
                return self.raw_identifier, accession, accession, True
            else:
                return self.raw_identifier, accession, 'NaN', True

    @unsync
    async def fetch_sequence(self, newest: bool = True):
        if self.source == 'RefSeq':
            res = await EutilsAPI.single_retrieve('efetch.fcgi',
                dict(db='sequences', id=self.identifier if newest else self.raw_identifier, rettype='fasta'),
                self.seq_folder['RefSeq'],
                self.eutils_api_web_semaphore)
            if res is not None:
                self.status = True
                return await a_seq_reader(res)
            else:
                self.status = False
                self.logger.warning(f'Invalid Identifier!')
        
        elif self.source == 'Ensembl':
            if self.status is None:
                await self.set_status()
            if self.status is False:
                self.logger.warning(f'Invalid Identifier!')
                return
            elif self.status['is_current'] != '1':
                self.logger.warning(
                    f'Not exists in current archive: \n{self.status}')
                return
            if not newest:
                self.logger.warning(
                    "Can't retrieve older version Ensembl Sequence via Ensembl REST API!")
            return await EnsemblAPI.single_retrieve('sequence/id/',
                self.identifier,
                dict(type='protein'),
                self.seq_folder['Ensembl'],
                self.ensembl_api_web_semaphore).then(a_seq_reader)

    @unsync
    async def map2unp(self, **kwargs):
        if self.level == 'model' or self.source == 'UniProt':
            return
        try:
            res = await self.map2unp_from_DB()
        except AssertionError:
            res = None
        if res is None:
            val = await self.fetch_from_ProteinsAPI(**kwargs)
            if val is None:
                return
            else:
                res = await self.map2unp_from_DB()
        return res


class Identifiers(tuple):
    def __new__(cls, iterable: Iterable):
        return super(Identifiers, cls).__new__(cls, (Identifier(i) if isinstance(i, str) else i for i in iterable))

    def __getitem__(self, slice):
        if isinstance(slice, str):
            for i in self:
                if i.identifier == slice:
                    return i
        res = tuple.__getitem__(self, slice)
        if isinstance(res, Iterable):
            return self.__class__(res)
        else:
            return res

    def fetch(self, func: Union[Callable[[Identifier], List], str], **kwargs):
        if isinstance(func, str):
            self.tasks = [getattr(id_ob, func)(**kwargs) for id_ob in self]
        else:
            self.tasks = [func(id_ob, **kwargs) for id_ob in self]
        return self
    
    @unsync
    async def run(self, tqdm=None):
        if tqdm is None:
            return [await fob for fob in as_completed(self.tasks)]
        else:
            return [await fob for fob in tqdm(as_completed(self.tasks), total=len(self.tasks))]

    def query_from_DB_with_unps(self, table_name: str, columns: str = '*'):
        default_tables = ('DB_REFERENCES', 'OTHER_DB_REFERENCES', 'ALTERNATIVE_PRODUCTS', 'FEATURES', 'INTERACTION', 'INFO')
        assert table_name in default_tables
        obs = tuple(i for i in self if i.source == 'UniProt')
        if len(obs) == 0:
            self.tasks = []
            return self
        accessions = tuple(i.identifier for i in obs)
        if columns != '*' and table_name == 'INFO':
            task = Identifier.sqlite_api.database.fetch_all(query=f'SELECT {columns} FROM INFO WHERE accession IN {accessions}')
        else:
            task = Identifier.sqlite_api.INFO.objects.filter(accession__in=accessions).all()
        exists = unsync_run(task)
        if len(exists) == 0:
            self.tasks = [ob.query_from_DB_with_unp(table_name=table_name, columns=columns, exists=False) for ob in obs]
            return self
        else:
            exist_ids = frozenset(i.accession for i in exists)
            rest_ids = frozenset(accessions) - exist_ids
            rest_dfs = [self[accession].query_from_DB_with_unp(table_name=table_name, columns=columns, exists=False) for accession in rest_ids]
            if table_name == 'INFO':
                ap = unsync_wrap(exists)
            else:
                if columns == '*':
                    ap = unsync_wrap(getattr(Identifier.sqlite_api, table_name).objects.filter(accession__in=exist_ids).all())
                else:
                    ap = unsync_wrap(Identifier.sqlite_api.database.fetch_all(query=f'SELECT {columns} FROM {table_name} WHERE accession IN {tuple(exist_ids)}'))
            rest_dfs.append(ap)
            self.tasks = rest_dfs
            return self



# @Created Date: 2020-09-17 12:02:31 am
# @Filename: __init__.py
# @Email:  1730416009@stu.suda.edu.cn
# @Author: ZeFeng Zhu
# @Last Modified: 2020-09-27 03:18:19 pm
# @Copyright (c) 2020 MinghuiGroup, Soochow University
import orm
from unsync import unsync
from pdb_profiling.processors.database import SqliteDB


class ProteinsDB(SqliteDB):

    def init_table_model(self):
        class ALTERNATIVE_PRODUCTS(orm.Model):
            __tablename__ = 'ALTERNATIVE_PRODUCTS'
            __metadata__ = self.metadata
            __database__ = self.database
            isoform = orm.String(max_length=50, primary_key=True)
            name = orm.JSON(allow_null=True)
            synonyms = orm.JSON(allow_null=True)
            sequenceStatus = orm.String(max_length=50)
            VAR_SEQ = orm.JSON(allow_null=True)
            accession = orm.String(max_length=50, primary_key=True)
            else_iso = orm.JSON(allow_null=True)
            iso_range = orm.JSON(allow_null=True)
            iso_range_len = orm.Integer(allow_null=True)
            sequence = orm.Text(allow_null=True)
            length = orm.Integer(allow_null=True)
        
        class INTERACTION(orm.Model):
            __tablename__ = 'INTERACTION'
            __metadata__ = self.metadata
            __database__ = self.database
            accession1 = orm.String(max_length=100, primary_key=True)
            accession2 = orm.String(max_length=100, primary_key=True)
            gene = orm.String(max_length=100, allow_null=True, allow_blank=True)
            interactor1 = orm.String(max_length=100, primary_key=True)
            interactor2 = orm.String(max_length=100, primary_key=True)
            organismDiffer = orm.Boolean()
            experiments = orm.Integer()
            chain1 = orm.Text(allow_null=True, allow_blank=True)
            chain2 = orm.Text(allow_null=True, allow_blank=True)
            accession = orm.String(max_length=50, primary_key=True)

        class DB_REFERENCES(orm.Model):
            __tablename__ = 'dbReferences'
            __metadata__ = self.metadata
            __database__ = self.database
            type = orm.String(max_length=100)
            isoform = orm.String(max_length=50, allow_null=True, allow_blank=True)
            accession = orm.String(max_length=50, primary_key=True)
            protein = orm.String(max_length=50, primary_key=True)
            transcript = orm.String(max_length=50)
            gene = orm.String(max_length=50, allow_null=True, allow_blank=True)

        class OTHER_DB_REFERENCES(orm.Model):
            __tablename__ = 'other_dbReferences'
            __metadata__ = self.metadata
            __database__ = self.database
            type = orm.String(primary_key=True, max_length=100)
       	    id = orm.String(primary_key=True, max_length=100)
            properties = orm.JSON(allow_null=True)
            isoform = orm.String(primary_key=True, max_length=100, allow_blank=True)
            evidences = orm.JSON(allow_null=True)
            accession = orm.String(primary_key=True, max_length=100)

        class FEATURES(orm.Model):
            __tablename__ = 'features'
            __metadata__ = self.metadata
            __database__ = self.database
            type = orm.String(primary_key=True, max_length=100)
            category = orm.String(primary_key=True, max_length=100)
            description = orm.Text(allow_null=True, allow_blank=True)
            begin = orm.Integer(primary_key=True)
            end = orm.Integer(primary_key=True)
            molecule = orm.String(primary_key=True, max_length=100, allow_blank=True)
            ftId = orm.String(max_length=100, allow_null=True, allow_blank=True)
            evidences = orm.JSON(allow_null=True)
            alternativeSequence = orm.Text(allow_null=True, allow_blank=True)
            accession = orm.String(primary_key=True, max_length=100)

        class INFO(orm.Model):
            __tablename__ = 'INFO'
            __metadata__ = self.metadata
            __database__ = self.database
            accession = orm.String(max_length=100, primary_key=True)
            id = orm.String(max_length=100)
            proteinExistence = orm.Text(allow_null=True, allow_blank=True)
            info = orm.JSON()
            organism = orm.JSON()
            secondaryAccession = orm.JSON(allow_null=True)
            protein = orm.JSON()
            gene = orm.JSON()
            sequence = orm.Text()
            length = orm.Integer()

        self.DB_REFERENCES = DB_REFERENCES
        self.OTHER_DB_REFERENCES = OTHER_DB_REFERENCES
        self.ALTERNATIVE_PRODUCTS = ALTERNATIVE_PRODUCTS
        self.FEATURES = FEATURES
        self.INTERACTION = INTERACTION
        self.INFO = INFO



# @Created Date: 2020-06-07 06:38:00 pm
# @Filename: select.py
# @Email:  1730416009@stu.suda.edu.cn
# @Author: ZeFeng Zhu
# @Last Modified: 2020-06-07 06:38:04 pm
# @Copyright (c) 2020 MinghuiGroup, Soochow University
from pandas import DataFrame, read_csv, isna, merge
import numpy as np
from unsync import unsync
import asyncio
import aiofiles
import ujson as json
import orjson
from pathlib import Path
import networkx as nx
from tablib import Dataset
from typing import Dict
from textdistance import overlap
from pdb_profiling.processors.pdbe.sqlite_api import converters
from pdb_profiling.processors.pdbe.neo4j_api import SIFTS, slice_series, lyst2range, related_dataframe
from pdb_profiling.utils import pipe_out
import logging

def str2ord(string: str):
    if len(string) > 1:
        num = 0
        for i in string:
            num += ord(i)
        return num
    else:
        return ord(string)


class Select_API(object):

    json_cols = ('entity_chain_map', 'unp_entity_info',
            'entity_unp_info', 'del_entity_chain')
    usecols = ['UniProt', 'pdb_id', 'chain_id',
               'new_unp_range', 'entity_id', 'RAW_BS']

    def __init__(self, sqlite, neo4j, folder, oscutoff, logger, **kwargs):
        self.sqlite = sqlite
        self.neo4j = neo4j
        self.oligo_path = folder/'oligo.tsv'
        self.moho_path = folder/'select_moho.tsv'
        self.he_path = folder/'select_he.tsv'
        self.oscutoff = oscutoff
        self.logger = logger
        self.kwargs = kwargs

    @staticmethod
    def select_range(ranges, indexes, cutoff=0.2, skip_index=[]):
        select_index = []
        def unit(cur_index):
            if cur_index in skip_index:
                return
            cur_range = orjson.loads(ranges[cur_index])
            for selected in select_index:
                selected_range = orjson.loads(ranges[selected])
                score = overlap.similarity(lyst2range(cur_range),
                                   lyst2range(selected_range))
                if score > cutoff:
                    return
            select_index.append(cur_index)

        for index in indexes:
            unit(index)
        return select_index

    @staticmethod
    def check_range(focus_sifts_nda, cutoff):
        skip_index = []
        # checked_id = []
        for checking_index, record in enumerate(focus_sifts_nda):
            pdb_id = record[1]
            # if pdb_id not in checked_id:
            #     checked_id.append(pdb_id)
            # else:
            #     continue
            related_records = focus_sifts_nda[focus_sifts_nda[:, 1] == pdb_id]
            if len(related_records) > 1:
                cur_range = orjson.loads(record[3])
                cur_chain = record[2]
                count = 0
                for other_record in related_records:
                    other_chain = other_record[2]
                    if other_chain == cur_chain:
                        continue
                    other_range = orjson.loads(other_record[3])
                    score = overlap.similarity(lyst2range(cur_range),
                                                lyst2range(other_range))
                    if score < 1 - cutoff:
                        continue
                    else:
                        count += 1
                if not count:
                    skip_index.append(checking_index)
            else:
                skip_index.append(checking_index)
        return skip_index

    def select_o(self, sifts_df, entry_info, cutoff=0.2, check_range:bool=False, tag:str='mo') -> Dataset:
        sifts_dict = slice_series(sifts_df.UniProt.to_numpy())
        sifts_nda = sifts_df.to_numpy()
        dataob = Dataset(headers=self.usecols+['1/resolution', 'revision_date_score', 'chain_id_score', 'select_tag', 'oligo_type'])
        for start, end in sifts_dict.values():
            focus_sifts_nda = sifts_nda[start:end]
            # TODO: check range
            if check_range:
                skip_index = self.check_range(focus_sifts_nda, cutoff)
            else:
                skip_index = []
            # TODO: Sort ndarray
            # Score, 1/R, Date, chain_id
            bs_chainid = focus_sifts_nda[:, (-1,2)]
            rs_date = [entry_info[pdb_id] for pdb_id in focus_sifts_nda[:, 1]]
            lyst = np.array([(bs, rs, str2ord(date), -str2ord(chain_id)) for (bs, chain_id), (rs, date) in zip(bs_chainid, rs_date)])
            sorted_index = np.lexsort(np.transpose(lyst)[::-1])[::-1]
            # TODO: select range
            selected_index = self.select_range(focus_sifts_nda[:, 3], sorted_index, cutoff, skip_index)
            new_sifts_nda = np.concatenate(
                (focus_sifts_nda[sorted_index, :-1], lyst[sorted_index]), axis=1)
            select_tag = np.array([(0, tag) for _ in range(len(sorted_index))])
            if len(selected_index):
                select_tag[selected_index] = np.array([(1, tag) for _ in range(len(selected_index))])
            new_sifts_nda = np.concatenate((new_sifts_nda, select_tag[sorted_index]), axis=1)
            for i in new_sifts_nda:
                dataob.append(list(i))
        return dataob

    @unsync
    async def select_e(self, sifts_df, entry_info, oligo_info, cutoff=0.2) -> Dataset:
        sifts_dict = slice_series(sifts_df.UniProt.to_numpy())
        sifts_nda = sifts_df.to_numpy()
        pass

    @staticmethod
    def yieldallchains(dictob):
        for lyst in dictob.values():
            yield from lyst
    
    @unsync
    async def process(self, sifts_path):

        def score_resolution(resolution):
            if resolution is None:
                return -1
            else:
                return 1/resolution

        sifts_df = read_csv(sifts_path, sep="\t", converters=converters, usecols=self.usecols)[self.usecols]
        sifts_df.drop_duplicates(inplace=True)
        sifts_df.sort_values(by='UniProt', inplace=True)
        sifts_dict = slice_series(sifts_df.UniProt.to_numpy())
        
        @unsync
        async def unit(sifts_df):
            pdbs = sifts_df.pdb_id.unique()
            oligo_df = await SIFTS.pipe_oligo(pdbs, self.neo4j.neo4j_api, self.sqlite, **self.kwargs)
            valid_chains = dict(zip(oligo_df.pdb_id, oligo_df.entity_chain_map.apply(
                lambda x: list(self.yieldallchains(x)))))
            check_chains = sifts_df[['pdb_id', 'chain_id']].apply(lambda x: x['chain_id'] in valid_chains[x['pdb_id']], axis=1)
            sifts_df.drop(index=check_chains[check_chains.eq(False)].index, inplace=True)
            pdbs = sifts_df.pdb_id.unique()
            entry_info = await self.sqlite.Entry_Info.objects.filter(pdb_id__in=pdbs).all()
            entry_info = {value.pdb_id: (score_resolution(value.resolution), value.REVISION_DATE) for value in entry_info}
            # TODO: select
            # dropna
            pdbs_focus = oligo_df.dropna(subset=['oligo_state']).pdb_id
            sifts_df = sifts_df[sifts_df.pdb_id.isin(pdbs_focus)]
            # mo
            pdbs_mo = oligo_df[oligo_df.oligo_state.eq('mo')].pdb_id
            sifts_mo_df = sifts_df[sifts_df.pdb_id.isin(pdbs_mo)]
            if len(sifts_mo_df) > 0:
                mo_res = self.select_o(sifts_mo_df, entry_info, self.oscutoff)
            else:
                mo_res = None
            # ho (+he)
            sifts_others_df = sifts_df.loc[sifts_df.index.difference(sifts_mo_df.index)]
            if len(sifts_others_df) > 0:
                ho_res = self.select_o(sifts_others_df, entry_info, self.oscutoff, True, 'ho')
            else:
                ho_res = None
            # he
            pdbs_he = oligo_df[oligo_df.oligo_state.eq('he')].pdb_id.to_numpy()
            sifts_he_df = sifts_others_df[sifts_others_df.pdb_id.isin(pdbs_he)]
            ## TODO: Partner Pipe
            oligo_info = {key: value for key,value in zip(oligo_df.pdb_id, oligo_df.entity_unp_info) if key in pdbs_he}
            ### -----------------------------------------------------------------
            new_sifts_df = await self.neo4j.pipe_new_sifts(pdbs_he, 'pdb')
            ### -----------------------------------------------------------------
            # TODO: output
            for col in self.json_cols:
                oligo_df[col] = oligo_df[col].apply(lambda x: json.dumps(x) if not isna(x) else x)
            await pipe_out(oligo_df, self.oligo_path)
            if mo_res is not None and ho_res is not None:
                await pipe_out(mo_res.stack(ho_res), self.moho_path)
            elif mo_res is not None and ho_res is None:
                await pipe_out(mo_res, self.moho_path)
            elif mo_res is None and ho_res is not None:
                await pipe_out(ho_res, self.moho_path)
            else:
                pass
            if new_sifts_df is not None:
                await pipe_out(new_sifts_df, self.he_path)

        tasks = asyncio.as_completed([unit(sifts_df.iloc[start:end].reset_index(drop=True)) for start, end in sifts_dict.values()])
        [await task for task in tasks]




# @Created Date: 2019-11-24 09:07:03 pm
# @Filename: setup.py
# @Email:  1730416009@stu.suda.edu.cn
# @Author: ZeFeng Zhu
# @Last Modified: 2019-12-23 04:27:14 pm
# @Copyright (c) 2019 MinghuiGroup, Soochow University
from setuptools import setup, find_packages, find_namespace_packages
import pdb_profiling

with open("README.md", "rt") as f:
    readme = f.read()


setup(
    name="pdb_profiling",
    version=pdb_profiling.__version__,

    packages=find_namespace_packages(),
    entry_points={'console_scripts': ['pdb_profiling=pdb_profiling.commands.command:Interface']},
    install_requires=[
        'aiosqlite>=0.13.0',
        'aioftp>=0.15.0',
        'aiohttp>=3.6.2',
        'aiofiles>=0.4.0',
        'unsync>=1.2.1',
        'tenacity>=4.12.0',
        'furl>=2.1.0',
        'orjson>=3.0.2',
        'pyexcel>=0.6.4',
        'tablib>=1.0.0',
        'pandas>=1.0.3',
        'numpy>=1.18.1',
        'textdistance>=4.1.5',
        'databases>=0.3.2',
        'neo4j>=4.0.1',
        'tqdm>=4.43.0',
        'orm>=0.1.5',
        'smart_open>=1.9.0',
        'scipy>=1.4.1',
        'scikit-learn>=0.22',
        'python-slugify>=4.0.0',
        'cachetools>=4.1.0'
     ],
    license="MIT",
    author_email="1730416009@stu.suda.edu.cn",
    maintainer="ZeFeng Zhu",
    maintainer_email="1730416009@stu.suda.edu.cn",
    description="Profiling Protein Structures from Protein Data Bank and integrate various resources.🏄‍♂️",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/NatureGeorge/pdb-profiling",
    python_requires=">=3.6.*",
    classifiers=[
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9"
    ],
    )

"""
Packaged By
python setup.py sdist bdist_wheel
twine upload --repository-url https://test.pypi.org/legacy/ dist/*
twine upload --repository-url https://upload.pypi.org/legacy/ dist/*

> https://packaging.python.org/tutorials/packaging-projects/
"""



# @Created Date: 2020-01-12 01:27:18 pm
# @Filename: api.py
# @Email:  1730416009@stu.suda.edu.cn
# @Author: ZeFeng Zhu
# @Last Modified: 2020-02-11 04:22:22 pm
# @Copyright (c) 2020 MinghuiGroup, Soochow University
import os
from numpy import array, nan, count_nonzero
import pandas as pd
from tablib import Dataset, InvalidDimensions, UnsupportedFormat
from typing import Union, Optional, Iterator, Iterable, Dict, List, Any, Generator, Callable, Tuple
from json import JSONDecodeError
import orjson as json
from pathlib import Path
from aiofiles import open as aiofiles_open
from collections import defaultdict
from unsync import unsync, Unfuture
from random import choice
from pdb_profiling.processors.pdbe import default_id_tag
from pdb_profiling.utils import decompression, related_dataframe, flatten_dict, pipe_out, dumpsParams
from pdb_profiling.log import Abclog
from pdb_profiling.fetcher.webfetch import UnsyncFetch
from pdb_profiling.processors.transformer import Dict2Tabular
from pdb_profiling.exceptions import WithoutExpectedKeyError

BASE_URL: str = 'https://www.ebi.ac.uk/pdbe/'

FTP_URL: str = 'ftp://ftp.ebi.ac.uk/'

FTP_DEFAULT_PATH: str = 'pub/databases/msd/sifts/flatfiles/tsv/uniprot_pdb.tsv.gz'

PDB_ARCHIVE_URL_EBI: str = 'http://ftp.ebi.ac.uk/pub/databases/pdb/data/structures/'
PDB_ARCHIVE_URL_WWPDB: str = 'https://ftp.wwpdb.org/pub/pdb/data/structures/'
PDB_ARCHIVE_VERSIONED_URL: str = 'http://ftp-versioned.wwpdb.org/pdb_versioned/data/'

# https://ftp.wwpdb.org/pub/pdb/data/structures/obsolete/mmCIF/a0/2a01.cif.gz
# http://ftp.ebi.ac.uk/pub/databases/pdb/data/structures/obsolete/mmCIF/a0/2a01.cif.gz
# http://ftp-versioned.wwpdb.org/pdb_versioned/data/entries/wm/pdb_00002wmg/pdb_00002wmg_xyz_v1-2.cif.gz

FUNCS = list()


def str_number_converter(x):
    try:
        return int(x)
    except ValueError:
        return -1


def dispatch_on_set(*keys):
    '''
    Decorator to add new dispatch functions
    '''
    def register(func):
        FUNCS.append((func, frozenset(keys)))
        return func
    return register


def traverseSuffixes(query: Any, *args):
    for func, keySet in FUNCS:
        if query in keySet:
            return func(*args)
    else:
        raise ValueError(f'Invalid query: {query}')


class ProcessPDBe(Abclog):

    converters = {
        'pdb_id': str,
        'chain_id': str,
        'struct_asym_id': str,
        'entity_id': str_number_converter,
        'author_residue_number': int,
        'residue_number': str_number_converter,
        'author_insertion_code': str,
        'id': int,
        'interface_id': int,
        'interface_number': int,
        'pdb_code': str,
        'assemble_code': int,
        'assembly_id': int,
        'oper_expression': str,
        'structure_1.range': str,
        'structure_2.range': str
        }
    
    use_existing: bool = False

    @staticmethod
    def yieldTasks(pdbs: Union[Iterable, Iterator], suffix: str, method: str, folder: str, chunksize: int = 25, task_id: int = 0) -> Generator:
        file_prefix = suffix.replace('/', '%')
        method = method.lower()
        if method == 'post':
            url = f'{BASE_URL}{suffix}'
            for i in range(0, len(pdbs), chunksize):
                params = {'url': url, 'data': ','.join(pdbs[i:i+chunksize])}
                yield method, params, os.path.join(folder, f'{file_prefix}+{task_id}+{i}.json')
        elif method == 'get':
            for pdb in pdbs:
                # pdb = pdb.lower()
                identifier = pdb.replace('/', '%')
                yield method, {'url': f'{BASE_URL}{suffix}{pdb}'}, os.path.join(folder, f'{file_prefix}+{identifier}.json')
        else:
            raise ValueError(f'Invalid method: {method}, method should either be "get" or "post"')

    @classmethod
    def single_retrieve(cls, pdb: str, suffix: str, method: str, folder: Union[Path, str], semaphore, rate: float = 1.5, **kwargs):
        return UnsyncFetch.single_task(
            task=next(cls.yieldTasks((pdb, ), suffix, method, folder)),
            semaphore=semaphore,
            to_do_func=kwargs.get('to_do_func', cls.process),
            rate=rate)

    @classmethod
    def retrieve(cls, pdbs: Union[Iterable, Iterator], suffix: str, method: str, folder: str, chunksize: int = 20, concur_req: int = 20, rate: float = 1.5, task_id: int = 0, ret_res:bool=True, **kwargs):
        # t0 = time.perf_counter()
        res = UnsyncFetch.multi_tasks(
            cls.yieldTasks(pdbs, suffix, method, folder, chunksize, task_id), 
            cls.process, 
            concur_req=concur_req, 
            rate=rate, 
            ret_res=ret_res,
            semaphore=kwargs.get('semaphore', None))
        # elapsed = time.perf_counter() - t0
        # cls.logger.info('{} ids downloaded in {:.2f}s'.format(len(res), elapsed))
        return res
    
    @classmethod
    @unsync
    async def process(cls, path: Union[str, Path, Unfuture]):
        cls.logger.debug('Start to decode')
        if not isinstance(path, (str, Path)):
            path = await path # .result()
        if path is None:
            return path
        path = Path(path)
        suffix = path.name.replace('%', '/').split('+')[0]
        new_path = Path(str(path).replace('.json', '.tsv'))
        if new_path.exists() and cls.use_existing and (new_path.stat().st_size > 0):
            return new_path
        async with aiofiles_open(path) as inFile:
            try:
                data = json.loads(await inFile.read())
            except Exception as e:
                cls.logger.error(f"Error in {path}")
                raise e
        if new_path.exists() and cls.use_existing and (new_path.stat().st_size > 0):
            return new_path
        res = Dict2Tabular.pyexcel_io(traverseSuffixes(suffix, data))
        if res is not None:
            if isinstance(res, Generator):
                count = 0
                for r in res:
                    if r is not None:
                        await pipe_out(df=r, path=new_path, format='tsv', mode='a' if count else 'w')
                        count += 1
                if not count:
                    cls.logger.debug(f"Without Expected Data ({suffix}): {data}")
                    return None
            else:
                await pipe_out(df=res, path=new_path, format='tsv', mode='w')
            cls.logger.debug(f'Decoded file in {new_path}')
            return new_path
        else:
            cls.logger.debug(f"Without Expected Data ({suffix}): {data}")
            return None


class PDBeDecoder(object):

    @staticmethod
    @dispatch_on_set('api/pdb/entry/status/', 'api/pdb/entry/summary/', 'api/pdb/entry/modified_AA_or_NA/',
                      'api/pdb/entry/mutated_AA_or_NA/', 'api/pdb/entry/cofactor/', 'api/pdb/entry/molecules/',
                      'api/pdb/entry/ligand_monomers/', 'api/pdb/entry/experiment/', 'api/pdb/entry/carbohydrate_polymer/',
                      'api/pdb/entry/electron_density_statistics/', 'api/pdb/entry/related_experiment_data/',
                      'api/pdb/entry/drugbank/', 'api/mappings/best_structures/',
                      'graph-api/pdb/mutated_AA_or_NA/', 'graph-api/pdb/modified_AA_or_NA/',
                      'graph-api/mappings/best_structures/', 'graph-api/compound/atoms/',
                      'graph-api/compound/bonds/', 'graph-api/compound/summary/',
                      'graph-api/compound/cofactors/', 'graph-api/pdb/funpdbe/',
                      'graph-api/pdb/bound_excluding_branched/',
                      'graph-api/pdb/bound_molecules/', 'graph-api/pdb/ligand_monomers/')
    def yieldCommon(data: Dict) -> Generator:
        for pdb in data:
            values = data[pdb]
            for value in values:
                for key in value:
                    if isinstance(value[key], (Dict, List)):
                        value[key] = json.dumps(value[key]).decode('utf-8')
            yield values, (default_id_tag(pdb, '_code_'),), (pdb,)

    @staticmethod
    @dispatch_on_set('api/pdb/entry/polymer_coverage/')
    def yieldPolymerCoverage(data: Dict) -> Generator:
        for pdb in data:
            molecules = data[pdb]['molecules']
            for entity in molecules:
                chains = entity['chains']
                for chain in chains:
                    observed = chain['observed']
                    for fragement in observed:
                        for key in ('start', 'end'):
                            fragement[key] = json.dumps(fragement[key]).decode('utf-8')
                    yield observed, ('chain_id', 'struct_asym_id', 'entity_id', 'pdb_id'), (chain['chain_id'], chain['struct_asym_id'], entity['entity_id'], pdb)

    @staticmethod
    @dispatch_on_set('api/pdb/entry/observed_residues_ratio/')
    def yieldObservedResiduesRatio(data: Dict) -> Generator:
        for pdb in data:
            for entity_id, entity in data[pdb].items():
                yield entity, ('entity_id', 'pdb_id'), (entity_id, pdb)

    @staticmethod
    @dispatch_on_set('api/pdb/entry/residue_listing/')
    def yieldResidues(data: Dict) -> Generator:
        for pdb in data:
            molecules = data[pdb]['molecules']
            for entity in molecules:
                chains = entity['chains']
                for chain in chains:
                    residues = chain['residues']
                    for res in residues:
                        if 'multiple_conformers' not in res:
                            res['multiple_conformers'] = ''
                        else:
                            res['multiple_conformers'] = json.dumps(res['multiple_conformers']).decode('utf-8')
                    yield residues, ('chain_id', 'struct_asym_id', 'entity_id', 'pdb_id'), (chain['chain_id'], chain['struct_asym_id'], entity['entity_id'], pdb)

    @staticmethod
    @dispatch_on_set('api/pdb/entry/secondary_structure/', 'graph-api/pdb/secondary_structure/')
    def yieldSecondaryStructure(data: Dict) -> Generator:
        for pdb in data:
            molecules = data[pdb]['molecules']
            for entity in molecules:
                chains = entity['chains']
                for chain in chains:
                    secondary_structure = chain['secondary_structure']
                    for name in secondary_structure:
                        fragment = secondary_structure[name]
                        for record in fragment:
                            for key in record:
                                if isinstance(record[key], (Dict, List)):
                                    record[key] = json.dumps(record[key]).decode('utf-8')
                            if 'sheet_id' not in record:
                                record['sheet_id'] = None
                        yield fragment, ('secondary_structure', 'chain_id', 'struct_asym_id', 'entity_id', 'pdb_id'), (name, chain['chain_id'], chain['struct_asym_id'], entity['entity_id'], pdb)

    @staticmethod
    @dispatch_on_set('api/pdb/entry/binding_sites/')
    def yieldBindingSites(data: Dict) -> Generator:
        for pdb in data:
            for site in data[pdb]:
                for tage in ('site_residues', 'ligand_residues'):
                    residues = site[tage]
                    for res in residues:
                        if 'symmetry_symbol' not in res:
                            res['symmetry_symbol'] = None
                    yield residues, ('residues_type', 'details', 'evidence_code', 'site_id', 'pdb_id'), (tage, site['details'], site['evidence_code'], site['site_id'], pdb)

    @staticmethod
    @dispatch_on_set('api/pdb/entry/assembly/')
    def yieldAssembly(data: Dict) -> Generator:
        for pdb in data:
            for biounit in data[pdb]:
                entities = biounit['entities']
                for entity in entities:
                    for key in entity:
                        if isinstance(entity[key], (Dict, List)):
                            entity[key] = json.dumps(entity[key]).decode('utf-8')
                keys = list(biounit)
                keys.remove('entities')
                yield entities, tuple(keys)+('pdb_id',), tuple(biounit[key] for key in keys)+(pdb, )

    @staticmethod
    @dispatch_on_set('api/pdb/entry/files/')
    def yieldAssociatedFiles(data: Dict) -> Generator:
        for pdb in data:
            for key in data[pdb]:
                for innerKey in data[pdb][key]:
                    record = data[pdb][key][innerKey]
                    if record:
                        yield record, ('innerKey', 'key', 'pdb_id'), (innerKey, key, pdb)
                    else:
                        continue

    @staticmethod
    @dispatch_on_set('api/mappings/all_isoforms/', 'api/mappings/uniprot/',
                      'api/mappings/uniprot_segments/', 'api/mappings/isoforms/',
                      'api/mappings/uniref90/', 'api/mappings/homologene_uniref90/',
                      'api/mappings/interpro/', 'api/mappings/pfam/',
                      'api/mappings/cath/', 'api/mappings/cath_b/',
                      'api/mappings/scop/', 'api/mappings/go/',
                      'api/mappings/ec/', 'api/mappings/ensembl/',
                      'api/mappings/hmmer/', 'api/mappings/sequence_domains/',
                      'api/mappings/structural_domains/', 'api/mappings/homologene/',
                      'api/mappings/uniprot_to_pfam/', 'api/mappings/uniprot_publications/',
                      'graph-api/mappings/uniprot/', 'graph-api/mappings/uniprot_segments/',
                      'graph-api/mappings/all_isoforms/', 'graph-api/mappings/',
                      'graph-api/mappings/isoforms/', 'graph-api/mappings/ensembl/',
                      'graph-api/mappings/homologene/', 'graph-api/mappings/sequence_domains/',
                      'api/mappings/'
                      # 'graph-api/uniprot/'
                      )
    def yieldSIFTSAnnotation(data: Dict) -> Generator:
        valid_annotation_set = {'UniProt', 'Ensembl', 'Pfam', 'CATH', 'CATH-B', 'SCOP', 'InterPro', 'GO', 'EC', 'Homologene', 'HMMER'}
        for top_root in data:
            # top_root: PDB_ID or else ID
            if data[top_root].keys() <= valid_annotation_set:
                # from PDB to ('UniProt', 'Ensembl', 'Pfam', 'CATH', 'CATH-B', 'SCOP', 'InterPro', 'GO', 'EC', 'Homologene', 'HMMER')
                # from PDB_ENTITY (i.e. graph-api/mappings/homologene/)
                # OR: from Uniprot (i.e. api/mappings/uniprot_to_pfam/)
                for sec_root in data[top_root]:
                    child = data[top_root][sec_root]
                    for annotation in child:
                        chains = child[annotation]['mappings']
                        for chain in chains:
                            for key, value in chain.items():
                                chain[key] = json.dumps(value).decode('utf-8') if isinstance(value, Dict) else value
                            for key, value in child[annotation].items():
                                if key == 'mappings':
                                    continue
                                chain[key] = json.dumps(value).decode('utf-8') if isinstance(value, Dict) else value
                            chain[default_id_tag(top_root, raise_error=True)] = top_root
                            chain[sec_root] = annotation
                        yield chains, None
            elif len(data[top_root].keys()) == 1 and 'PDB' in data[top_root].keys():
                # from UniProt to PDB
                for sec_root in data[top_root]:
                    child = data[top_root][sec_root]
                    for pdb in child:
                        chains = child[pdb]
                        for chain in chains:
                            chain['start'] = json.dumps(chain['start']).decode('utf-8')
                            chain['end'] = json.dumps(chain['end']).decode('utf-8')
                        yield chains, ('pdb_id', 'UniProt'), (pdb, top_root)
            else:
                raise ValueError(f'Unexpected data structure for inputted data: {data}')

    @staticmethod
    @dispatch_on_set('api/pisa/interfacelist/')
    def yieldPISAInterfaceList(data: Dict):
        for pdb in data:
            try:
                records = data[pdb]['interfaceentries']
            except KeyError:
                raise WithoutExpectedKeyError(f"Without Expected interface_detail: {data}")
            for record in records:
                flatten_dict(record, 'structure_1')
                flatten_dict(record, 'structure_2')
            flatten_dict(data[pdb], 'page_title', False)
            cols = sorted(i for i in data[pdb].keys() if i != 'interfaceentries')
            yield records, cols, tuple(data[pdb][col] for col in cols)
    
    @staticmethod
    @dispatch_on_set('api/pisa/interfacedetail/')
    def yieldPISAInterfaceDetail(data: Dict):
        usecols = (
            'pdb_code', 'assemble_code', 'interface_number',
            'interface_detail.interface_structure_1.structure.selection',
            'interface_detail.interface_structure_2.structure.selection')
        edge_cols1 = ('structure',)  # 'interface_atoms', 'interface_residue', 'interface_area', 'solvation_energy'
        edge_cols2 = ('structure',)  # 'interface_atoms', 'interface_residues', 'interface_area', 'solvation_energy'
        for pdb in data:
            try:
                records = data[pdb]['interface_detail']
            except KeyError:
                raise WithoutExpectedKeyError(f"Without Expected interface_detail: {data}")
            del records['bonds']
            for col in edge_cols1: flatten_dict(records['interface_structure_1'], col)
            for col in edge_cols2: flatten_dict(records['interface_structure_2'], col)
            flatten_dict(data[pdb], 'page_title', False)
            flatten_dict(records, 'interface_structure_1')
            flatten_dict(records, 'interface_structure_2')
            flatten_dict(data[pdb], 'interface_detail')
            # cols = sorted(i for i in data[pdb].keys() if i != 'interface_detail.residues')
            yield data[pdb]['interface_detail.residues']['residue1']['residue']['residue_array'], usecols, tuple(data[pdb][col] for col in usecols)
            yield data[pdb]['interface_detail.residues']['residue2']['residue']['residue_array'], usecols, tuple(data[pdb][col] for col in usecols)

    @staticmethod
    @dispatch_on_set('graph-api/residue_mapping/')
    def graph_api_residue_mapping(data: Dict):
        '''
        * <https://www.ebi.ac.uk/pdbe/graph-api/residue_mapping/:pdbId/:entityId/:residueNumber>
        * <https://www.ebi.ac.uk/pdbe/graph-api/residue_mapping/:pdbId/:entityId/:residueStart/:residueEnd>

        NOTE: only yield UniProt Residue Related Data
        '''
        cols = (
            'pdb_id', 'entity_id', 'chain_id', 'struct_asym_id',
            'residue_number', 'author_residue_number',
            'author_insertion_code', 'observed', 'UniProt')

        for pdb_id in data:
            assert len(data[pdb_id]) == 1, f"Unexpected Cases: {pdb_id}"
            molecules = data[pdb_id][0]
            for chain in molecules['chains']:
                for residue in chain['residues']:
                    yield list({**dict(zip(cols, (
                        pdb_id, molecules['entity_id'], chain['auth_asym_id'],
                        chain['struct_asym_id'], residue['residue_number'],
                        residue['author_residue_number'], residue['author_insertion_code'],
                        residue['observed'], feature_tag))), **feature} for feature_tag, feature in residue['features']['UniProt'].items()), None

    @staticmethod
    @dispatch_on_set('graph-api/pdb/sequence_conservation/')
    def sequence_conservation(data: Dict):
        for pdb in data:
            yield [{
                    'pdb_id': pdb, 
                    'entity_id': data[pdb]['entity_id'],
                    'length': data[pdb]['length'],
                    'residue_number': val['start'],
                    'conservation_score': val['conservation_score'],
                    'letter_array': json.dumps(tuple(i['letter'] for i in val['amino'])).decode('utf-8'),
                    'proba_array': json.dumps(tuple(i['proba'] for i in val['amino'])).decode('utf-8')}
                    for val in data[pdb]['data']], None
            # letter_array, proba_array = zip(*((i['letter'], i['proba']) for i in val['amino']))

    @staticmethod
    @dispatch_on_set('graph-api/pdb/funpdbe_annotation/depth/',
                     'graph-api/pdb/funpdbe_annotation/cath-funsites/',
                     'graph-api/pdb/funpdbe_annotation/3Dcomplex/',
                     'graph-api/pdb/funpdbe_annotation/akid/',
                     'graph-api/pdb/funpdbe_annotation/3dligandsite/',
                     'graph-api/pdb/funpdbe_annotation/camkinet/',
                     'graph-api/pdb/funpdbe_annotation/canSAR/',
                     'graph-api/pdb/funpdbe_annotation/ChannelsDB/',
                     'graph-api/pdb/funpdbe_annotation/dynamine/',
                     'graph-api/pdb/funpdbe_annotation/FoldX/',
                     'graph-api/pdb/funpdbe_annotation/MetalPDB/',
                     'graph-api/pdb/funpdbe_annotation/M-CSA/',
                     'graph-api/pdb/funpdbe_annotation/p2rank/',
                     'graph-api/pdb/funpdbe_annotation/Missense3D/',
                     'graph-api/pdb/funpdbe_annotation/POPScomp_PDBML/',
                     'graph-api/pdb/funpdbe_annotation/ProKinO/',
                     'graph-api/pdb/funpdbe_annotation/14-3-3-pred/',
                     'graph-api/pdb/funpdbe_annotation/'
                     )
    def funpdbe_resources(data: Dict):
        for pdb in data:
            info = data[pdb]
            for val in info:
                for annotation in val['annotations']:
                    yield annotation['site_residues'], ('pdb_id', 'origin', 'evidence_codes', 'site_id', 'label'), (pdb, val['origin'], val['evidence_codes'], annotation['site_id'], annotation['label'])

    @staticmethod
    @dispatch_on_set('graph-api/pdbe_pages/rfam/',
                     'graph-api/pdbe_pages/annotations/',
                     'graph-api/pdbe_pages/uniprot_mapping/',
                     'graph-api/pdbe_pages/binding_sites/',
                     'graph-api/pdbe_pages/interfaces/',
                     'graph-api/pdbe_pages/secondary_structure/',
                     'graph-api/pdbe_pages/domains/',
                     'graph-api/uniprot/unipdb/',
                     'graph-api/uniprot/annotations/',
                     'graph-api/uniprot/interface_residues/',
                     'graph-api/uniprot/ligand_sites/',
                     'graph-api/uniprot/secondary_structures/',
                     'graph-api/uniprot/domains/',
                     'graph-api/uniprot/sequence_conservation/')
    def graph_api_data_common(data: Dict):
        for pdb in data:
            for info in data[pdb]['data']:
                if 'additionalData' in info:
                    flatten_dict(info, 'additionalData')
                com_keys = tuple(key for key in info.keys() if key != 'residues')
                yield info['residues'], ('pdb_id',)+com_keys, (pdb,)+tuple(info[key] for key in com_keys)


class PDBeModelServer(Abclog):
    '''
    Implement ModelServer API
    '''
    
    root = f'{BASE_URL}model-server/v1/'
    headers =  {'accept': 'text/plain', 'Content-Type': 'application/json'}
    api_set = frozenset(('atoms', 'residueInteraction', 'assembly', 'full', 'ligand'
                'residueSurroundings', 'symmetryMates', 'query-many'))
    
    @classmethod
    def yieldTasks(cls, pdbs, suffix: str, method: str, folder: str, data_collection, params, filename='subset') -> Generator:
        if data_collection is None:
            assert method == 'get', 'Invalid method!'
            for pdb in pdbs:
                args = dict(
                    url=f'{cls.root}{pdb}/{suffix}?',
                    headers=cls.headers,
                    params=params)
                yield method, args, os.path.join(folder, f'{pdb}_{filename}.{params.get("encoding", "cif")}')
        else:
            assert method == 'post', 'Invalid method!'
            for pdb, data in zip(pdbs, data_collection):
                args = dict(
                    url=f'{cls.root}{pdb}/{suffix}?',
                    headers=cls.headers,
                    params=params,
                    data=data)
                yield method, args, os.path.join(folder, f'{pdb}_{filename}.{params.get("encoding", "cif")}')

    @classmethod
    def retrieve(cls, pdbs, suffix: str, method: str, folder: str, data_collection=None, params=None, concur_req: int = 20, rate: float = 1.5, ret_res:bool=True, **kwargs):
        if params is None:
            params = {'model_nums': 1, 'encoding': 'cif'}
        res = UnsyncFetch.multi_tasks(
            cls.yieldTasks(pdbs, suffix, method, folder,
                           data_collection, params, kwargs.get('filename', 'subset')),
            concur_req=concur_req,
            rate=rate,
            ret_res=ret_res,
            semaphore=kwargs.get('semaphore', None))
        return res
    
    @classmethod
    def single_retrieve(cls, pdb: str, suffix: str, method: str, folder: Union[Path, str], semaphore, data_collection=None, params=None, rate: float = 1.5, filename='subset'):
        if params is None:
            params = {'encoding': 'cif'}
        if data_collection is not None:
            data_collection = (data_collection, )
        return UnsyncFetch.single_task(
            task=next(cls.yieldTasks((pdb, ), suffix, method, folder,
                                     data_collection, params, filename=filename)),
            semaphore=semaphore,
            rate=rate)


class PDBeCoordinateServer(Abclog):

    roots = (f'{BASE_URL}coordinates/', 'https://cs.litemol.org/')
    headers = {'accept': 'text/plain'}
    api_set = frozenset(('ambientResidues', 'assembly', 'backbone', 'cartoon', 'chains'
                         'entities', 'full', 'het', 'ligandInteraction', 'residueRange',
                         'residues', 'sidechain', 'symmetryMates', 'trace', 'water'))

    def __init__(self, root: str = 'random'):
        if root == 'random':
            self.root = choice(self.roots)
        elif root == 'ebi':
            self.root = self.roots[0]
        elif root == 'litemol':
            self.root = self.roots[1]
        else:
            raise ValueError("root should be (ebi, litemol, random)")
    
    def __repr__(self):
        return f'<CoordinateServerAPI: {self.root}>'
    
    def task_unit(self, pdb_id, suffix: str, params, folder):
        args = dict(
                url=f'{self.root}{pdb_id}/{suffix}?',
                headers=self.headers,
                params=params)
        return 'get', args, Path(folder)/f'{pdb_id}_{dumpsParams(params)}.{params.get("encoding", "cif")}'
    
    def single_retrieve(self, pdb_id: str, suffix: str, params:Dict, folder: Union[Path, str], semaphore, rate: float = 1.5):
        return UnsyncFetch.single_task(
            task=self.task_unit(pdb_id, suffix, params, folder),
            semaphore=semaphore,
            rate=rate)


class PDBArchive(Abclog):
    '''
    Download files from PDB Archive

    * wwPDB/RCSB: PDB_ARCHIVE_URL_WWPDB: str = 'https://ftp.wwpdb.org/pub/pdb/data/structures/'
    * EBI: PDB_ARCHIVE_URL_EBI: str = 'http://ftp.ebi.ac.uk/pub/databases/pdb/data/structures/'
    '''
    root = PDB_ARCHIVE_URL_EBI
    api_set = frozenset(f'{i}/{j}/' for i in ('obsolete', 'divided')
                for j in ('mmCIF', 'pdb', 'XML'))

    @classmethod
    def task_unit(cls, pdb: str, suffix: str, file_suffix: str, folder: Path):
        args = dict(url=f'{cls.root}{suffix}{pdb[1:3]}/{pdb}{file_suffix}')
        return 'get', args, folder/f'{pdb}{file_suffix}'

    @classmethod
    def yieldTasks(cls, pdbs, suffix: str, file_suffix: str, folder: Path) -> Generator:
        for pdb in pdbs:
            yield cls.task_unit(pdb, suffix, file_suffix, folder)

    @classmethod
    def retrieve(cls, pdbs, suffix: str, folder: Path, file_suffix: str = '.cif.gz', concur_req: int = 20, rate: float = 1.5, ret_res:bool=True, **kwargs):
        res = UnsyncFetch.multi_tasks(
            cls.yieldTasks(pdbs, suffix, file_suffix, folder),
            concur_req=concur_req,
            rate=rate,
            ret_res=ret_res,
            semaphore=kwargs.get('semaphore', None))
        return res
    
    @classmethod
    def single_retrieve(cls, pdb, suffix: str, folder: Path, semaphore, file_suffix: str = '.cif.gz', rate: float = 1.5):
        return UnsyncFetch.single_task(
            task=cls.task_unit(pdb, suffix, file_suffix, folder),
            semaphore=semaphore,
            rate=rate)


class PDBVersioned(PDBArchive):
    '''
    Download files from PDB Versioned

    * wwPDB Versioned: PDB_ARCHIVE_VERSIONED_URL: str = 'http://ftp-versioned.wwpdb.org/pdb_versioned/data/entries/'

    >>> PDBVersioned.single_retrieve(
        ('2wmg', '_v1-2'), 'entries/', 
        init_folder_from_suffix(Base.get_folder(), 'pdb-versioned/entries'), 
        Base.get_web_semaphore()).result()
    '''
    root = PDB_ARCHIVE_VERSIONED_URL
    api_set = frozenset(('entries/', 'removed/'))

    @classmethod
    def task_unit(cls, pdb_with_version: Tuple, suffix: str, file_suffix: str, folder: Path):
        pdb, version_info = pdb_with_version
        file_name = f'pdb_0000{pdb}_xyz{version_info}{file_suffix}'
        args = dict(url=f'{cls.root}{suffix}{pdb[1:3]}/pdb_0000{pdb}/{file_name}')
        return 'get', args, folder/file_name




# @Created Date: 2020-05-25 10:21:37 pm
# @Filename: dbfetch.py
# @Email:  1730416009@stu.suda.edu.cn
# @Author: ZeFeng Zhu
# @Last Modified: 2020-05-26 12:17:51 am
# @Copyright (c) 2020 MinghuiGroup, Soochow University
from __future__ import absolute_import
import asyncio
from unsync import unsync, Unfuture
from neo4j import GraphDatabase
from neo4j.exceptions import ServiceUnavailable
from typing import Callable, List, Optional
from pandas import DataFrame, concat

"""
Currently support DB:

* Neo4j
* Sqlite
"""


class Neo4j:
    """
    Neo4j database API

    Add support for asyncio/unsync to neo4j-python-driver (Based on savv's `Neo4j Class`)

    Reference
    ---------
    * <https://github.com/neo4j/neo4j-python-driver/issues/180#issuecomment-380056816>
    """

    # How long to wait after each successive failure.
    RETRY_WAITS: List = [0, 1, 4]

    def __init__(self, config, concur_req: int = 200, semaphore = None, log_func: Callable = print):
        self._config = config
        self.log_func = log_func
        if semaphore is None:
            self._semaphore = asyncio.Semaphore(concur_req)
        else:
            self.log_func("Neo4j.init: pass asyncio.Semaphore")
            self._semaphore = semaphore
        
    
    @unsync
    async def connnect(self) -> Unfuture:
        for retry_wait in self.RETRY_WAITS:
            try:
                await self.init_driver()
                break
            except Exception as e:
                if retry_wait == self.RETRY_WAITS[-1]:
                    raise
                else:
                    self.log_func(f'!!!: retrying to Init DB; err: {e}')
                    # wait for 0, 1, 3... seconds.
                    await asyncio.sleep(retry_wait)
        return self

    @unsync
    def init_driver(self):
        self.log_func("Init Neo4j DataBase Driver")
        self.driver = GraphDatabase.driver(self._config['url'], auth=(self._config['user'], self._config['pass']))

    @unsync
    def afetch_start(self, query, **kwargs):
        with self.driver.session() as session:
            return [dict(i) for i in session.run(query, **kwargs)]
    
    @unsync
    async def afetch(self, query, **kwargs):
        for retry_wait in self.RETRY_WAITS:
            try:
                async with self._semaphore:
                    res = await self.afetch_start(query, **kwargs)
                break
            except (BrokenPipeError, ServiceUnavailable) as e:
                if retry_wait == self.RETRY_WAITS[-1]:
                    raise e
                else:
                    self.log_func('BrokenPipeError or ServiceUnavailable. Retrying...')
                    await asyncio.sleep(retry_wait)
                    await self.init_driver()
        return res
         
    def close(self):
        """Close Neo4j DataBase Driver"""
        self.log_func("Close Neo4j DataBase Driver")
        self.driver.close()


if __name__ == "__main__":
    import sys
    sys.path.append("C:/GitWorks/pdb-profiling")
    from pdb_profiling.processors.pdbe.neo4j_api import Entry
    config = {'user': '...', 'pass': '...',
              'url': 'bolt://...'}
    queries = [
        Entry.get_residues('1a01', '1', 'A', observed_only=True),
        Entry.get_residues('2xyn', '1', 'A', observed_only=True),
        Entry.get_residues('1z8g', '1', 'A', observed_only=True)
        ]
    demo = Neo4j(config).connnect().result()
    ress = [demo.afetch(query, **kwargs).result() for query, kwargs in queries]
    result = concat((DataFrame(dict(i) for i in res) for res in ress), ignore_index=True, sort=False)
    print(result)
    demo.close()



# @Created Date: 2020-01-12 01:27:18 pm
# @Filename: transformer.py
# @Email:  1730416009@stu.suda.edu.cn
# @Author: ZeFeng Zhu
# @Last Modified: 2020-02-11 04:22:22 pm
# @Copyright (c) 2020 MinghuiGroup, Soochow University
from pyexcel import get_sheet, Sheet
from tablib import Dataset
from pandas import DataFrame, concat
from collections import OrderedDict
from typing import Dict, Iterable, Tuple


class Dict2Tabular(object):
    """
    Convert Json-Dict to Tabular format
    """

    @staticmethod
    def sync_with_pandas(*args) -> DataFrame:
        records, *remain = args
        if not len(records):
            return None
        df = DataFrame(records)
        if len(remain) > 1:
            for append_header, append_value in zip(*remain):
                df[append_header] = append_value
        return df

    @classmethod
    def pandas_io(cls, data: Iterable[Tuple]):
        return concat((cls.sync_with_pandas(*res) for res in data), ignore_index=True, sort=False)

    @staticmethod
    def sync_with_pyexcel(*args) -> Sheet:
        records, *remain = args
        if not len(records):
            return None
        keys = {key for record in records for key in record.keys()}
        head = records[0]
        records[0] = {key: head.get(key, None) for key in keys}
        sheet = get_sheet(records=records)
        if len(remain) > 1:
            append_header, append_value = remain
            append_value = tuple(i if i is not None else '' for i in append_value)
            append_data = [append_header] + [append_value for _ in range(len(records))]
            sheet.column += Sheet(append_data)
        sheet.name_columns_by_row(0)
        return sheet

    @classmethod
    def pyexcel_io(cls, data: Iterable[Tuple]) -> Sheet:
        cur_sheet = None
        new_sheets = []
        for res in data:  # traverseSuffixes(suffix, data)
            if cur_sheet is None:
                cur_sheet = cls.sync_with_pyexcel(*res)
                if cur_sheet is None:
                    continue
                cur_cols = set(cur_sheet.colnames)
            else:
                new_sheet = cls.sync_with_pyexcel(*res)
                if new_sheet is None:
                    continue
                new_cols = set(new_sheet.colnames)
                # assert cur_cols >= new_cols, f"Unexpected new columns: {new_cols-cur_cols}"
                if not cur_cols >= new_cols:
                    append_header = tuple(new_cols-cur_cols)
                    append_data = [append_header] + [
                        tuple('' for _ in range(len(append_header))) for _ in range(len(cur_sheet))]
                    # NOTE: cur_sheet and new_sheet have colnames while the new Sheet does not
                    cur_sheet.column += Sheet(append_data)
                    cur_cols |= new_cols
                new_sheets.append(new_sheet)
                for sheet in new_sheets:
                    e_cols = set(sheet.colnames)
                    x_cols = cur_cols-e_cols
                    if len(x_cols) > 0:
                        append_header = tuple(x_cols)
                        append_data = [append_header] + [
                            tuple('' for _ in range(len(append_header))) for _ in range(len(sheet))]
                        sheet.column += Sheet(append_data)
        yield cur_sheet
        yield from new_sheets

    @staticmethod
    def sync_with_tablib(*args) -> Dataset:
        records, *remain = args
        if not len(records):
            return None
        keys = {key for record in records for key in record.keys()}
        ob = Dataset()
        records = [OrderedDict(sorted((key, record.get(key, None)) for key in keys))
                   for record in records]
        ob.dict = records
        if len(remain) > 1:
            append_header, append_value = remain
            for i in range(len(append_value)):
                ob.append_col([append_value[i]]*len(records), append_header[i])
        return ob

    @classmethod
    def tablib_io(cls, data: Iterable[Tuple]) -> Dataset:
        cur_ob = None
        for res in data:
            try:
                cur_cols = set(cur_ob.headers)
                new = cls.sync_with_tablib(*res)
                new_cols = set(new.headers)
                assert cur_cols == new_cols, f"Unequal columns: new_cols-cur_cols={new_cols-cur_cols}, cur_cols-new_cols={cur_cols-new_cols}"
                cur_ob.dict += new.dict
            except AttributeError:
                cur_ob = cls.sync_with_tablib(*res)
        return cur_ob



# @Created Date: 2020-04-08 09:44:01 am
# @Filename: neo4j_api.py
# @Email:  1730416009@stu.suda.edu.cn
# @Author: ZeFeng Zhu
# @Last Modified: 2020-04-08 09:44:05 am
# @Copyright (c) 2020 MinghuiGroup, Soochow University
from typing import List, Iterable, Iterator, Union, Dict, Optional, Tuple
from functools import partial, reduce
import pandas as pd
from pathlib import Path
from collections import defaultdict
from itertools import combinations, product
from numpy import nan
import numpy as np
import orjson as json
from functools import lru_cache
from textdistance import jaccard, overlap
from Bio import Align
from Bio.SubsMat import MatrixInfo as matlist
from unsync import unsync, Unfuture
import asyncio
import aiofiles
from tablib import Dataset
import traceback
from pdb_profiling.utils import (pipe_out, sort_sub_cols, slice_series, to_interval, 
                                 lyst22intervel, SEQ_DICT, standardAA, standardNu, range_len,
                                 interval2set, lyst2range, subtract_range,
                                 add_range, overlap_range, outside_range_len, related_dataframe)
from pdb_profiling.log import Abclog
from pdb_profiling.fetcher.dbfetch import Neo4j
from pdb_profiling.processors.pdbe.sqlite_api import Sqlite_API
import logging
# logging.basicConfig(level=logging.INFO)


def lyst2dict(lyst: List) -> Dict:
    try:
        res = dict(lyst)
    except TypeError:
        res = dict((tuple(key), value)for key,value in lyst)
    for key in res.keys():
        res[key] = dict(res[key])
    return res


def sub_index(init_index, subtract_index) -> pd.Index:
    if len(subtract_index) == 0:
        return init_index
    else:
        return pd.Index(set(init_index)-set(subtract_index))


def subtract_dict(left: Dict, right: Dict, copy:bool=True) -> Dict:
    if copy:
        left = left.copy()
    for key in left.keys() & right.keys():
        res = set(left[key]) - set(right[key])
        if res:
            left[key] = list(res)
        else:
            del left[key]
    return left


def tidy_na(dfrm: pd.DataFrame, colName: str, fill, astype):
    dfrm[colName] = dfrm[colName].fillna(fill).apply(astype)


class Entry(object):

    session = None
    
    @classmethod
    def set_session(cls, session):
        cls.session = session
    
    @staticmethod
    def to_data_frame(res):
        return pd.DataFrame(dict(i) for i in res)
    
    @classmethod
    def summary_method(cls, pdbs, session=None):
        query = '''
            MATCH (entry:Entry)-[:EXPERIMENT]->(method:Method)
            WHERE entry.ID in $lyst
            RETURN entry.ID as pdb_id,
                   entry.PDB_REV_DATE_ORIGINAL as PDB_REV_DATE_ORIGINAL, 
                   entry.FIRST_REV_DATE as FIRST_REV_DATE,
                   entry.PDB_REV_DATE as PDB_REV_DATE,
                   entry.REVISION_DATE as REVISION_DATE,
                   tofloat(entry.RESOLUTION) as resolution,
                   method.METHOD_CLASS as METHOD_CLASS
        '''
        if session is None:
            if cls.session is not None:
                session = cls.session
        try:
            return session.run(query, lyst=list(pdbs))
        except AttributeError:
            return query, dict(lyst=list(pdbs))

    @classmethod
    def summary_ligand(cls, pdbs, session=None):
        query = '''
            MATCH (entry:Entry)-[:HAS_BOUND_MOLECULE]->(bmol:BoundMolecule)<-[:IS_PART_OF]-(bli:BoundLigand) 
            WHERE entry.ID in $lyst
            RETURN entry.ID as pdb_id, COUNT(distinct bli) as BOUND_LIGAND_COUNT, COUNT(distinct bmol) as BOUND_MOL_COUNT
        '''
        if session is None:
            if cls.session is not None:
                session = cls.session
        try:
            return session.run(query, lyst=list(pdbs))
        except AttributeError:
            return query, dict(lyst=list(pdbs))
    
    @classmethod
    def summary_nucleotides(cls, pdbs, session=None):
        query = '''
            MATCH (entry:Entry)-[:HAS_ENTITY]->(entity:Entity)-[:HAS_PDB_RESIDUE]->(pdbRes:PDBResidue)
            WHERE entity.POLYMER_TYPE IN ['D','R','D/R'] AND entry.ID in $lyst
            RETURN 
                entry.ID as pdb_id, 
                entity.ID as entity_id, 
                entity.POLYMER_TYPE as polymer_type,
                size([x in collect(pdbRes.CHEM_COMP_ID) where x in ['DA','DT','DC','DG']]) as DNA_COUNT,
                size([x in collect(pdbRes.CHEM_COMP_ID) where x in ['A','U','C','G']]) as RNA_COUNT,
                size([x in collect(pdbRes.CHEM_COMP_ID) where not x in $standardNu]) as OTHER_COUNT
        '''
        if session is None:
            if cls.session is not None:
                session = cls.session
        try:
            return session.run(query, lyst=list(pdbs), standardNu=standardNu)
        except AttributeError:
            return query, dict(lyst=list(pdbs), standardNu=standardNu)
    
    @staticmethod
    def get_polymer_type(dna, rna):
        if dna and rna:
            return 'D/R'
        elif dna and not rna:
            return 'D'
        elif not dna and rna:
            return 'R'
        elif not dna and not rna:
            return nan

    @classmethod
    def deal_nucleotides(cls, res):
        if isinstance(res, pd.DataFrame):
            dfrm = res
        elif res is None:
            return None
        else:
            dfrm = cls.to_data_frame(res)
        if len(dfrm) == 0:
            return None
        # dfrm['polymer_type'] = dfrm.apply(lambda x: cls.get_polymer_type(x['DNA_COUNT'], x['RNA_COUNT']), axis=1)
        dfrm['polymer_type'] = [cls.get_polymer_type(x, y) for x, y in zip(dfrm['DNA_COUNT'], dfrm['RNA_COUNT'])]
        res = pd.DataFrame(
            ((pdb_id, json.dumps(dict(zip(data.entity_id, data.polymer_type))).decode('utf-8'))
             for pdb_id, data in dfrm.groupby('pdb_id')),
            columns=('pdb_id', 'nucleotides_entity_type'))
        # res['has_hybrid_nucleotides'] = res.nucleotides_entity_type.apply(lambda x: 'D\/R' in x)
        res['has_hybrid_nucleotides'] = ['D/R' in x for x in res.nucleotides_entity_type.values]
        return res

    @classmethod
    def summary_seq(cls, pdbs, session=None):
        query = '''
            MATCH (entry:Entry)-[:HAS_ENTITY]->(entity:Entity{POLYMER_TYPE:'P'})-[:CONTAINS_CHAIN]->(chain:Chain)-[inChain:IS_IN_CHAIN]-(res:PDBResidue)
            WHERE entry.ID in $lyst
            WITH 
                entry.ID as pdb_id,
                entity.ID as entity_id,
                chain.AUTH_ASYM_ID as chain_id,
                COUNT(res) as SEQRES_COUNT,
                avg(tofloat(inChain.OBSERVED_RATIO)) as AVG_OBS_RATIO,
                [x in COLLECT([res.ID, inChain.AUTH_COMP_ID]) WHERE NOT x[1] IN $standardAA] as NON_INDEX,
                [x in COLLECT([res.ID, inChain.AUTH_COMP_ID]) WHERE x[1] = 'UNK'] as UNK_INDEX,
                [x in COLLECT([res.ID, inChain.OBSERVED]) WHERE x[1]='N'] as MIS_INDEX,
                chain
            MATCH (chain)-[obs_inChain:IS_IN_CHAIN]-()
           	WHERE tofloat(obs_inChain.OBSERVED_RATIO) > 0
            RETURN pdb_id, entity_id, chain_id, SEQRES_COUNT, AVG_OBS_RATIO, avg(tofloat(obs_inChain.OBSERVED_RATIO)) as AVG_OBS_OBS_RATIO, NON_INDEX, UNK_INDEX, MIS_INDEX
        '''
        if session is None:
            if cls.session is not None:
                session = cls.session
        try:
            return session.run(query, lyst=list(pdbs), standardAA=standardAA)
        except AttributeError:
            return query, dict(lyst=list(pdbs), standardAA=standardAA)

    @classmethod
    def summary_eec(cls, pdbs, protein_only:bool=True):
        query = '''
            MATCH (entry:Entry)-[:HAS_ENTITY]-(entity:Entity{})-[:CONTAINS_CHAIN|:IS_AN_INSTANCE_OF]-(chain)
            WHERE entry.ID IN $pdbs
            RETURN entry.ID as pdb_id, entity.ID as entity_id, chain.AUTH_ASYM_ID as chain_id, chain.STRUCT_ASYM_ID as struct_asym_id, entity.POLYMER_TYPE as POLYMER_TYPE
        '''
        if protein_only:
            query = query.format("{POLYMER_TYPE:'P'}")
        else:
            query = query.format("")
        return query, dict(pdbs=list(pdbs))

    @staticmethod
    def deal_seq_index(dfrm: pd.DataFrame) -> pd.DataFrame:
        def index2range(lyst):
            res = to_interval(int(i[0]) for i in lyst)
            if isinstance(res, List):
                return json.dumps(res).decode('utf-8')
            else:
                return res
        
        def get_obs_rel_count(seqres_count, mis_range, unk_range, non_range):
            seq_range = [(1, seqres_count)]
            obs_range = subtract_range(seq_range, mis_range)
            obs_unk_range = overlap_range(obs_range, unk_range)
            obs_pure_range = subtract_range(obs_range, non_range)
            return range_len(obs_unk_range), range_len(obs_pure_range)

        dfrm['UNK_COUNT'] = dfrm.UNK_INDEX.apply(len)
        dfrm['PURE_SEQRES_COUNT'] = dfrm.SEQRES_COUNT - dfrm.NON_INDEX.apply(len)
        dfrm['OBS_RECORD_COUNT'] = dfrm.SEQRES_COUNT - dfrm.MIS_INDEX.apply(len)
        dfrm.NON_INDEX = [index2range(x) for x in dfrm.NON_INDEX]
        dfrm.UNK_INDEX = [index2range(x) for x in dfrm.UNK_INDEX]
        dfrm.MIS_INDEX = [index2range(x) for x in dfrm.MIS_INDEX]
        dfrm[['OBS_UNK_COUNT', 'ATOM_RECORD_COUNT']] = dfrm.apply(lambda x: get_obs_rel_count(
            x['SEQRES_COUNT'], x['MIS_INDEX'], x['UNK_INDEX'], x['NON_INDEX']), axis=1, result_type='expand')
        return dfrm

    @classmethod
    @lru_cache()
    def get_seqres(cls, pdb_id: str, entity_id: str = None, chain_id: str = None, three2one: bool = True, session=None):
        '''
        TODO: Different behavior
        '''
        if session is None:
            if cls.session is not None:
                session = cls.session
        if entity_id is not None:
            query = '''
                MATCH (entity:Entity)-[:HAS_PDB_RESIDUE]->(pdbRes:PDBResidue)
                WHERE entity.UNIQID = $uniqid
                RETURN pdbRes.CHEM_COMP_ID AS residue_name ORDER BY toInteger(pdbRes.ID)
            '''
            # res = session.run(query, uniqid=f'{pdb_id}_{entity_id}')
            kwargs = dict(uniqid=f'{pdb_id}_{entity_id}')
        elif chain_id is not None:
            query = '''
                MATCH (entry:Entry)-[:HAS_ENTITY]-(:Entity)-[:CONTAINS_CHAIN]-(chain:Chain)-[:IS_IN_CHAIN]-(pdbRes:PDBResidue)
                WHERE entry.ID = $pdb_id AND chain.AUTH_ASYM_ID = $chain_id
                RETURN pdbRes.CHEM_COMP_ID AS residue_name ORDER BY toInteger(pdbRes.ID)
            '''
            # res = session.run(query, pdb_id=pdb_id, chain_id=chain_id)
            kwargs = dict(pdb_id=pdb_id, chain_id=chain_id)
        else:
            raise ValueError('please specify entity_id or chain_id')
        
        try:
            res = session.run(query, **kwargs)
            if three2one:
                return ''.join(SEQ_DICT.get(r['residue_name'], 'X') for r in res)
            else:
                return res
        except AttributeError:
            return query, kwargs

    @classmethod
    def get_residues(cls, pdb_id: str, entity_id: str, chain_id: str, res_ids=None, observed_only:bool=False, session=None):
        if session is None:
            if cls.session is not None:
                session = cls.session
        query = '''
            MATCH (entry:Entry)-[:HAS_ENTITY]->(entity:Entity)-[:CONTAINS_CHAIN]->(chain:Chain)-[inChain:IS_IN_CHAIN]-(res:PDBResidue)
            WHERE entry.ID = $pdb_id AND entity.ID = $entity_id AND chain.AUTH_ASYM_ID = $chain_id {}
            RETURN entry.ID as pdb_id, entity.ID as entity_id, chain.AUTH_ASYM_ID as chain_id, res.CHEM_COMP_ID as residue_name, toInteger(res.ID) as residue_number, tofloat(inChain.OBSERVED_RATIO) as obs_ratio, toInteger(inChain.AUTH_SEQ_ID) as author_residue_number, inChain.PDB_INS_CODE as author_insertion_code ORDER BY residue_number
        '''
        if res_ids is None:
            if observed_only:
                query = query.format("AND inChain.OBSERVED = 'Y'")
            else:
                query = query.format('')
            # return session.run(query, pdb_id=pdb_id, entity_id=str(entity_id), chain_id=chain_id)
            kwargs = dict(pdb_id=pdb_id, entity_id=str(entity_id), chain_id=chain_id)
        else:
            if observed_only:
                query = query.format("AND res.ID IN $res_ids AND inChain.OBSERVED = 'Y'")
            else:
                query = query.format('AND res.ID IN $res_ids')
            res_ids = [str(i) for i in res_ids]
            # return session.run(query, pdb_id=pdb_id, entity_id=str(entity_id), chain_id=chain_id, res_ids=res_ids)
            kwargs = dict(pdb_id=pdb_id, entity_id=str(entity_id), chain_id=chain_id, res_ids=res_ids)
        try:
            return session.run(query, **kwargs)
        except AttributeError:
            return query, kwargs

    @classmethod
    def get_binding_redidues(cls, pdb_id: str, entity_id: str, chain_id: str, count_only: bool = True, session=None):
        if session is None:
            if cls.session is not None:
                session = cls.session
        if count_only:
            query = '''
                MATCH (entry:Entry)-[:HAS_BOUND_MOLECULE]->(:BoundMolecule)<-[:IS_PART_OF]-(:BoundLigand)-[:HAS_ARP_CONTACT]-(res:PDBResidue)-[:IS_IN_CHAIN]-(chain:Chain)-[:CONTAINS_CHAIN]-(entity:Entity)
                WHERE entry.ID = $pdb_id AND entity.ID = $entity_id AND chain.AUTH_ASYM_ID = $chain_id
                RETURN COUNT(DISTINCT res) as LIGAND_BINDING_RES_COUNT
            '''
        else:
            query = '''
                MATCH (entry:Entry)-[:HAS_BOUND_MOLECULE]->(:BoundMolecule)<-[:IS_PART_OF]-(:BoundLigand)-[:HAS_ARP_CONTACT]-(res:PDBResidue)-[inChain:IS_IN_CHAIN]-(chain:Chain)-[:CONTAINS_CHAIN]-(entity:Entity)
                WHERE entry.ID = $pdb_id AND entity.ID = $entity_id AND chain.AUTH_ASYM_ID = $chain_id
                RETURN distinct entry.ID as pdb_id, 
                                entity.ID as entity_id, 
                                chain.AUTH_ASYM_ID as chain_id, 
                                res.CHEM_COMP_ID as residue_name, 
                                toInteger(res.ID) as residue_number, 
                                tofloat(inChain.OBSERVED_RATIO) as obs_ratio, 
                                toInteger(inChain.AUTH_SEQ_ID) as author_residue_number, 
                                inChain.PDB_INS_CODE as author_insertion_code
                '''
        kwargs = dict(
            pdb_id=pdb_id, 
            entity_id=str(entity_id), 
            chain_id=chain_id)
        try:
            return session.run(query, **kwargs)
        except AttributeError:
            return query, kwargs

    @classmethod
    def summary_entity_chain(cls, pdbs, session=None):
        query = """
                MATCH (entry:Entry)-[:HAS_ENTITY]-(entity:Entity{POLYMER_TYPE:'P'})-[:CONTAINS_CHAIN]-(chain:Chain)
                WHERE entry.ID IN $pdbs
                WITH entry, entity, COLLECT(chain.AUTH_ASYM_ID) as chain_ids
                RETURN entry.ID as pdb_id, COLLECT([entity.ID, chain_ids]) as entity_chain_map
            """
        if session is None:
            if cls.session is not None:
                session = cls.session
        try:
            return session.run(query, pdbs=list(pdbs))
        except AttributeError:
            return query, dict(pdbs=list(pdbs))
    
    @classmethod
    def deal_entity_chain(cls, res):
        if isinstance(res, pd.DataFrame):
            dfrm = res
        else:
            dfrm = cls.to_data_frame(res)
            dfrm.entity_chain_map = dfrm.entity_chain_map.apply(dict)
        dfrm['entity_count'] = dfrm.entity_chain_map.apply(len)
        dfrm['chain_count'] = dfrm.entity_chain_map.apply(lambda x: sum(len(i) for i in x.values()))
        return dfrm
 
    @classmethod
    def summary_entity_interaction(cls, pdbs, session=None):
        '''
        Target: 
        
        * Whether two chains have interaction, if yes, find the interacting residues;
        * Then check whether the interacting residues are in the mapped range of the corresponding uniprot-isoform

        Possible Solution:

        1. Find the interacting entity
        2. Fing the Interface Node
        3. Directly go to the residues with contact
        4. Use InterfaceBindingResidue Node
        5. Use MAP_TO_UNIPROT_RESIDUE Relationship

        Problem to solve:

        1. Difference between ARP_CONTACT and PISA_CONTACT
        2. 
        '''
        query = '''
                MATCH (entry:Entry)-[:HAS_ENTITY]-(entityA:Entity{POLYMER_TYPE:'P'})-[:INTERACTS_WITH_ARP|:INTERACTS_WITH_PISA]-(entityB:Entity{POLYMER_TYPE:'P'})
                WHERE entry.ID IN $pdbs
                RETURN entityA, entityB
            '''
        if session is None:
            if cls.session is not None:
                session = cls.session
        try:
            return session.run(query, pdbs=list(pdbs))
        except AttributeError:
            return query, dict(lyst=list(pdbs))

    @classmethod
    def summary_chain_interaction(cls, pdbs, source:str):
        query = '''
        MATCH (entry:Entry)-[:HAS_ENTITY]-(:Entity{POLYMER_TYPE:'P'})-[:HAS_PDB_RESIDUE]-(:PDBResidue)-[has_contact:%s]-(:PDBResidue)
        WHERE entry.ID IN $pdbs AND has_contact.AUTH_ASYM_ID_1 <> has_contact.AUTH_ASYM_ID_2
        RETURN distinct entry.ID as pdb_id, has_contact.AUTH_ASYM_ID_1 as chain_id_1, has_contact.AUTH_ASYM_ID_2 as chain_id_2, true as from_%s
        '''
        # //COLLECT(has_contact.TYPE) as type
        source = source.lower()
        if source == 'pisa_bond':
            query = query % ("HAS_PISA_BOND", source)
        elif source == 'arp_contact':
            query = query % ("HAS_ARP_CONTACT", source)
        else:
            raise ValueError("Invalid source! Please input specified source within ('pisa_bond', 'arp_contact')")
        return query, dict(pdbs=list(pdbs))

    @classmethod
    def summary_assembly(cls, pdbs, protein_only:bool=True):
        query = '''
        MATCH (entry:Entry)-[:HAS_ENTITY]-(entity:Entity{})-[:CONTAINS_CHAIN]-(chain:Chain)-[:IS_PART_OF_ASSEMBLY]-(ass:Assembly)
        WHERE entry.ID IN $pdbs
        WITH entry,entity,chain,ass
        MATCH (entity:Entity)-[inass:IS_PART_OF_ASSEMBLY]-(ass:Assembly)
        RETURN
            entry.ID as pdb_id,
            entity.ID as entity_id,
            entity.POLYMER_TYPE as POLYMER_TYPE,
            chain.AUTH_ASYM_ID as chain_id,
            ass.UNIQID as biounit_id, 
            ass.PREFERED as PREFERED,
            inass.NUMBER_OF_CHAINS as NUMBER_OF_CHAINS
        '''
        if protein_only:
            query = query.format("{POLYMER_TYPE:'P'}")
        else:
            query = query.format("")
        return query, dict(pdbs=list(pdbs))
    
    @classmethod
    def summary_assembly_interface(cls, pdbs, protein_only:bool=True):
        query = '''
        MATCH (entry:Entry)-[:HAS_ENTITY]-(entity:Entity{})-[:CONTAINS_CHAIN]-(chain:Chain)-[:IS_PART_OF_ASSEMBLY]-(ass:Assembly)-[:HAS_INTERFACE]-(interface:Interface)
        WHERE entry.ID IN $pdbs
        WITH entry,entity,chain,ass,interface
        MATCH (interface:Interface)-[:HAS_ENTITY]-(entity:Entity)-[inass:IS_PART_OF_ASSEMBLY]-(ass:Assembly)
        RETURN
            entry.ID as pdb_id,
            entity.ID as entity_id,
            entity.POLYMER_TYPE as POLYMER_TYPE,
            chain.AUTH_ASYM_ID as chain_id,
            ass.UNIQID as biounit_id, 
            ass.PREFERED as PREFERED,
            inass.NUMBER_OF_CHAINS as NUMBER_OF_CHAINS,
            interface.UNIQID as interface_id
        '''
        if protein_only:
            query = query.format("{POLYMER_TYPE:'P'}")
        else:
            query = query.format("")
        return query, dict(pdbs=list(pdbs))


class SIFTS(Entry):
    @classmethod
    def summary_entity_unp(cls, pdbs, tax_id: Optional[str]=None, session=None):
        query = """
                MATCH (entry:Entry)-[:HAS_ENTITY]-(entity:Entity{POLYMER_TYPE:'P'})-[unp_rel:HAS_UNIPROT{BEST_MAPPING:"1"}]-(unp:UniProt)-[:HAS_TAXONOMY]-(tax:Taxonomy)
                WHERE entry.ID in $pdbs %s
                WITH entry, entity, unp, tax
                MATCH (unp)-[seg:HAS_UNIPROT_SEGMENT]-(entity)
                WITH COLLECT(DISTINCT [toInteger(seg.UNP_START), toInteger(seg.UNP_END)]) as range_info, unp, entry, entity, tax
                WITH COLLECT([entity.ID, range_info]) as eneitiy_unp_info, entry, unp, tax
                RETURN entry.ID as pdb_id, COLLECT([[unp.ACCESSION,tax.TAX_ID], eneitiy_unp_info]) as unp_entity_info
            """
        if tax_id is None:
            query = query % ''
        else:
            query = query % f'AND tax.ID = "{tax_id}"'
        if session is None:
            if cls.session is not None:
                session = cls.session
        try:
            return session.run(query, pdbs=list(pdbs))
        except AttributeError:
            return query, dict(pdbs=list(pdbs))
    
    @classmethod
    def deal_entity_unp(cls, res):
        def reverse_dict(info_dict: Dict) -> Dict:
            res = defaultdict(list)
            for key, valueDict in info_dict.items():
                for innerKey in valueDict.keys():
                    res[innerKey].append(key)
            return res

        def min_unp(info_dict: Dict) -> int:
            return min(len(set(res)) for res in product(*info_dict.values()))
        
        if isinstance(res, pd.DataFrame):
            dfrm = res
        else:
            dfrm = cls.to_data_frame(res)
            dfrm.unp_entity_info = dfrm.unp_entity_info.apply(lambda x: lyst2dict(x) if not isinstance(x, float) else nan)
        # !!!
        dfrm['entity_unp_info'] = dfrm.unp_entity_info.apply(lambda x: reverse_dict(x) if not isinstance(x, float) else nan)
        dfrm['entity_with_unp_count'] = dfrm.entity_unp_info.apply(lambda x: len(x) if not isinstance(x, float) else nan)
        dfrm['min_unp_count'] = dfrm.entity_unp_info.apply(lambda x: min_unp(x) if not isinstance(x, float) else nan)
        return dfrm
    
    @staticmethod
    def get_oligo_state(df: pd.DataFrame):
        state_name, unmapped_name = 'oligo_state', 'has_unmapped_protein'
        assert len(df[df.entity_count < df.entity_with_unp_count]) == 0
        cur_df = df
        # He
        he = cur_df[
            ((cur_df.entity_count > cur_df.entity_with_unp_count) &
             (cur_df.min_unp_count.eq(1))) | (cur_df.min_unp_count.gt(1))
        ].index
        cur_df = cur_df.loc[sub_index(cur_df.index, he)]
        # Ho
        ho = cur_df[
            (cur_df.entity_count == cur_df.entity_with_unp_count)
            & (cur_df.min_unp_count.eq(1))
            & (cur_df.chain_count.gt(1))
        ].index
        cur_df = cur_df.loc[sub_index(cur_df.index, ho)]
        # Mo
        mo = df[df.chain_count.eq(1)].index
        cur_df_index = sub_index(cur_df.index, mo)
        # Tag
        df[state_name] = nan
        df.loc[mo, state_name] = 'mo'
        df.loc[ho, state_name] = 'ho'
        df.loc[he, state_name] = 'he'
        # Tag
        non_ho = df.loc[sub_index(df.index, ho)]
        unmapped = non_ho[non_ho.entity_count > non_ho.entity_with_unp_count].index
        df[unmapped_name] = False
        df.loc[unmapped, unmapped_name] = True
        # return
        return df, (mo, ho, he, cur_df_index)

    @classmethod
    def summary_oligo_state(cls, pdbs):
        entity_chain = cls.deal_entity_chain(cls.summary_entity_chain(pdbs))
        unp_entity = cls.deal_entity_unp(cls.summary_entity_unp(pdbs))
        unp_entity_chain = pd.merge(entity_chain, unp_entity, how='left')
        tidy_na(unp_entity_chain, 'entity_with_unp_count', 0, int)
        tidy_na(unp_entity_chain, 'min_unp_count', 0, int)
        # pass_result_oli, indexes_oli
        return cls.get_oligo_state(unp_entity_chain)
    
    @classmethod
    @unsync
    async def asummary_oligo_state(cls, pdbs, neo4j):
        query, kwargs = cls.summary_entity_chain(pdbs)
        res = await neo4j.afetch(query, **kwargs)
        entity_chain = cls.deal_entity_chain(res)
        query, kwargs = cls.summary_entity_unp(pdbs)
        res = await neo4j.afetch(query, **kwargs)
        unp_entity = cls.deal_entity_unp(res)
        unp_entity_chain = pd.merge(entity_chain, unp_entity, how='left')
        tidy_na(unp_entity_chain, 'entity_with_unp_count', 0, int)
        tidy_na(unp_entity_chain, 'min_unp_count', 0, int)
        return cls.get_oligo_state(unp_entity_chain)

    @classmethod
    @unsync
    async def pipe_oligo(cls, pdbs, neo4j, sqlite, **kwargs):
        seqres_df = await sqlite.SEQRES_Info.objects.filter(pdb_id__in=pdbs).all()
        seqres_df = pd.DataFrame(seqres_df)
        omit_df = cls.omit_chains(seqres_df, **kwargs)
        oligo_df, _ = await cls.asummary_oligo_state(pdbs, neo4j)
        oligo_df, _ = cls.update_oligo_state(oligo_df, omit_df)
        return oligo_df

    @staticmethod
    def omit_chains(seq_df: pd.DataFrame, cutoff: int = 50, omit_col: str = 'ATOM_RECORD_COUNT'):
        focus = seq_df[seq_df[omit_col].lt(cutoff)]
        return pd.DataFrame(
            ((pdb_id, dict((entity_id, entityData.chain_id.to_list()) for entity_id,
                           entityData in pdbData.groupby('entity_id'))) for pdb_id, pdbData in focus.groupby('pdb_id')),
            columns=('pdb_id', 'del_entity_chain')
        )

    @classmethod
    def update_oligo_state(cls, oligo_df: pd.DataFrame, omit_df: pd.DataFrame):
        def new_entity_chain_map(entity_chain_map, del_entity_chain) -> Dict:
            if isinstance(del_entity_chain, Dict):
                return subtract_dict(entity_chain_map, del_entity_chain)
            else:
                return entity_chain_map

        def rest_unp_entity(unp_entity_info, entity_chain_map: Dict):
            if not isinstance(unp_entity_info, Dict):
                return nan
            new_unp_entity_info = defaultdict(dict)
            for unp, valuedict in unp_entity_info.items():
                entities = valuedict.keys() & entity_chain_map.keys()
                if not entities:
                    continue
                for entity in entities:
                    new_unp_entity_info[unp][entity] = unp_entity_info[unp][entity]
            return new_unp_entity_info
        
        oligo_df = pd.merge(oligo_df, omit_df, how='left')
        oligo_df.entity_chain_map = oligo_df.apply(lambda x: new_entity_chain_map(
            x['entity_chain_map'], x['del_entity_chain']), axis=1)
        oligo_df.unp_entity_info = oligo_df.apply(lambda x: rest_unp_entity(
            x['unp_entity_info'], x['entity_chain_map']), axis=1)
        oligo_df = cls.deal_entity_unp(cls.deal_entity_chain(oligo_df))
        tidy_na(oligo_df, 'entity_with_unp_count', 0, int)
        tidy_na(oligo_df, 'min_unp_count', 0, int)
        return cls.get_oligo_state(oligo_df)

    @classmethod
    def set_from(cls, id_type: str):
        id_type = id_type.lower()
        if id_type in ('unp', 'uniprot', 'unp.ACCESSION'):
            cls.from_str = 'unp.ACCESSION'
        elif id_type in ('pdb', 'pdb_id', 'entry.ID'):
            cls.from_str = 'entry.ID'
        else:
            raise ValueError('unknown id type')

    @classmethod
    def summary_mapping(cls, lyst, id_type: str, session=None):
        cls.set_from(id_type)
        query = '''
            MATCH (entry:Entry)-[:HAS_ENTITY]-(entity:Entity)-[seg:HAS_UNIPROT_SEGMENT]-(unp:UniProt)
            WHERE {} in $lyst
            RETURN unp.ACCESSION as UniProt, entry.ID as pdb_id, entity.ID as entity_id, seg.AUTH_ASYM_ID as chain_id, tofloat(seg.IDENTITY) as identity, COLLECT([toInteger(seg.PDB_START), toInteger(seg.PDB_END)]) as pdb_range, COLLECT([toInteger(seg.UNP_START), toInteger(seg.UNP_END)]) as unp_range
        '''.format(cls.from_str)
        if session is None:
            if cls.session is not None:
                session = cls.session
        try:
            return session.run(query, lyst=list(lyst))
        except AttributeError:
            return query, dict(lyst=list(lyst))
    
    @classmethod
    def summary_mapping_detail(cls, pdb_id, entity_id, unp):
        query = '''
        MATCH (entry:Entry)-[:HAS_ENTITY]-(entity:Entity)-[:HAS_PDB_RESIDUE]-(pdbres:PDBResidue)-[:MAP_TO_UNIPROT_RESIDUE]-(unpres:UNPResidue)-[:HAS_UNP_RESIDUE]-(unp:UniProt)
        WHERE unp.ACCESSION = $unp AND entity.ID = $entity_id AND entry.ID = $pdb_id
        RETURN toInteger(pdbres.ID) as residue_number,
               toInteger(SPLIT(unpres.UNIQID, '_')[1]) as unp_index
        '''
        return query, dict(pdb_id=pdb_id, entity_id=entity_id, unp=unp)

    @classmethod
    @unsync
    async def deal_mapping(cls, res, neo4j_api):
        dfrm = cls.to_data_frame(res)
        if len(dfrm) == 0:
            return
        dfrm = cls.deal_InDe(dfrm)
        dfrm.pdb_range = dfrm.pdb_range.apply(
            lambda x: json.dumps(x).decode('utf-8'))
        dfrm.unp_range = dfrm.unp_range.apply(
            lambda x: json.dumps(x).decode('utf-8'))
        dfrm = await cls.update_range(dfrm, neo4j_api)
        # dfrm = await cls.update_range_local_align(dfrm, neo4j_api)
        return dfrm

    @classmethod
    def extra_mapping(cls, entity_uniqids, unps) -> pd.DataFrame:
        '''
        TODO: Different behavior
        '''
        query = '''
            MATCH (entity:Entity)-[seg:HAS_UNIPROT_SEGMENT]->(unp:UniProt)
            WHERE entity.UNIQID IN $entity_uniqids AND unp.ACCESSION IN $unps
            RETURN unp.ACCESSION as UniProt,
                   substring(entity.UNIQID,0,4) as pdb_id, 
                   entity.ID as entity_id, 
                   seg.AUTH_ASYM_ID as chain_id, 
                   tofloat(seg.IDENTITY) as identity, 
                   COLLECT([toInteger(seg.PDB_START), 
                   toInteger(seg.PDB_END)]) as pdb_range, 
                   COLLECT([toInteger(seg.UNP_START), 
                   toInteger(seg.UNP_END)]) as unp_range
        '''
        '''
        if session is None:
            if cls.session is not None:
                session = cls.session
        try:
            # TODO: this part of code is wrong right now, need to be fix
            res = session.run(query,
                              entity_uniqids=list(entity_uniqids),
                              unps=list(unps))
            return cls.deal_mapping(res)
        except AttributeError:
        '''
        return query, dict(entity_uniqids=list(entity_uniqids), unps=list(unps))

    @classmethod
    def pdbchain2unpres(cls, lyst, observed_only: bool = False):
        query = '''
        MATCH (chain:Chain)-[inChain:IS_IN_CHAIN]-(pdbres:PDBResidue)-[:MAP_TO_UNIPROT_RESIDUE]-(unpres:UNPResidue)-[:HAS_UNP_RESIDUE]-(unp:UniProt)
        WHERE chain.UNIQID in $lyst {}
        RETURN chain.UNIQID as UNIQID,
               chain.AUTH_ASYM_ID as chain_id,
               pdbres.CHEM_COMP_ID as residue_name, 
               toInteger(pdbres.ID) as residue_number,
               tofloat(inChain.OBSERVED_RATIO) as obs_ratio,
               toInteger(inChain.AUTH_SEQ_ID) as author_residue_number,
               inChain.PDB_INS_CODE as author_insertion_code,
               unpres.UNIQID as unp_index,
               unpres.ONE_LETTER_CODE as unp_res
        '''
        if observed_only:
            query = query.format("AND inChain.OBSERVED = 'Y'")
        else:
            query = query.format('')

        return query, dict(lyst=list(lyst))

    @staticmethod
    def sort_2_range(unp_range: List, pdb_range: List):
        unp_range, pdb_range = zip(*sorted(zip(unp_range, pdb_range), key=lambda x: x[0][0]))
        return unp_range, pdb_range

    @classmethod
    def deal_InDe(cls, dfrm: pd.DataFrame) -> pd.DataFrame:
        def get_gap_list(li: List):
            return [li[i+1][0] - li[i][1] - 1 for i in range(len(li)-1)]

        def get_range_diff(lyst_a: List, lyst_b: List):
            array_a = np.array([right - left + 1 for left, right in lyst_a])
            array_b = np.array([right - left + 1 for left, right in lyst_b])
            return (array_a - array_b).tolist()

        def add_tage_to_range(df: pd.DataFrame, tage_name: str):
            # ADD TAGE FOR SIFTS
            df[tage_name] = 'Safe'
            # No Insertion But Deletion[Pure Deletion]
            df.loc[df[(df['group_info'] == 1) & (
            df['unp_pdb_var'] > 0)].index, tage_name] = 'Deletion'
            # Insertion & No Deletion
            df.loc[df[
                (df['group_info'] != 1) &
                (df['var_0_count'] == df['group_info']) &
                (df['unp_gap_0_count'] == (df['group_info'] - 1))].index, tage_name] = 'Insertion'
            # Insertion & Deletion
            df.loc[df[
                (df['group_info'] != 1) &
                ((df['var_0_count'] != df['group_info']) |
                (df['unp_gap_0_count'] != (df['group_info'] - 1)))].index, tage_name] = 'Insertion & Deletion'

        dfrm['group_info'] = dfrm.apply(lambda x: len(
            x['pdb_range']), axis=1)
        
        focus_index = dfrm[dfrm.group_info.gt(1)].index
        if len(focus_index) > 0: 
            focus_df = dfrm.loc[focus_index].apply(lambda x: cls.sort_2_range(x['unp_range'], x['pdb_range']), axis=1, result_type='expand')
            focus_df.index = focus_index
            focus_df.columns = ['unp_range', 'pdb_range']
            dfrm.loc[focus_index, ['unp_range', 'pdb_range']] = focus_df
        
        dfrm['pdb_gap_list'] = dfrm.apply(lambda x: json.dumps(
            get_gap_list(x['pdb_range'])).decode('utf-8'), axis=1)
        dfrm['unp_gap_list'] = dfrm.apply(lambda x: json.dumps(
            get_gap_list(x['unp_range'])).decode('utf-8'), axis=1)
        dfrm['var_list'] = dfrm.apply(lambda x: json.dumps(get_range_diff(
            x['unp_range'], x['pdb_range'])).decode('utf-8'), axis=1)
        dfrm['repeated'] = dfrm.apply(
            lambda x: '-' in x['var_list'], axis=1)
        dfrm['repeated'] = dfrm.apply(
            lambda x: True if '-' in x['unp_gap_list'] else x['repeated'], axis=1)
        dfrm['var_0_count'] = dfrm.apply(
            lambda x: json.loads(x['var_list']).count(0), axis=1)
        dfrm['unp_gap_0_count'] = dfrm.apply(
            lambda x: json.loads(x['unp_gap_list']).count(0), axis=1)
        dfrm['unp_pdb_var'] = dfrm.apply(
            lambda x: json.loads(x['var_list'])[0], axis=1)
        add_tage_to_range(dfrm, tage_name='sifts_range_tag')
        return dfrm

    @classmethod
    @unsync
    async def update_range(cls, dfrm: pd.DataFrame, neo4j_api) -> pd.DataFrame:
        new_unp_range, new_pdb_range = 'new_unp_range', 'new_pdb_range'
        focus_df = dfrm[
            (dfrm.sifts_range_tag.isin(('Deletion', 'Insertion & Deletion')))
            & (dfrm.repeated.eq(False))]
        focus = focus_df[['pdb_id', 'entity_id', 'UniProt']].to_numpy()
        focus_index = focus_df.index
        updated_pdb_range, updated_unp_range = list(), list()
        for record in focus:
            query, kwargs = cls.summary_mapping_detail(*record)
            res = await neo4j_api.afetch(query, **kwargs)
            res = cls.to_data_frame(res)
            res_unp_intervel, res_pdb_intervel = lyst22intervel(res.unp_index, res.residue_number)
            updated_unp_range.append(json.dumps(res_unp_intervel).decode('utf-8'))
            updated_pdb_range.append(json.dumps(res_pdb_intervel).decode('utf-8'))
        updated_range_df = pd.DataFrame(
            {new_unp_range: updated_unp_range, new_pdb_range: updated_pdb_range}, index=focus_index)
        dfrm = pd.merge(dfrm, updated_range_df, left_index=True,
                        right_index=True, how='left')
        dfrm[new_unp_range] = dfrm.apply(lambda x: x['unp_range'] if pd.isna(
            x[new_unp_range]) else x[new_unp_range], axis=1)
        dfrm[new_pdb_range] = dfrm.apply(lambda x: x['pdb_range'] if pd.isna(
            x[new_pdb_range]) else x[new_pdb_range], axis=1)
        return dfrm

    @classmethod
    @unsync
    async def update_range_local_align(cls, dfrm: pd.DataFrame, neo4j_api) -> pd.DataFrame:
        new_unp_range, new_pdb_range = 'new_unp_range_align', 'new_pdb_range_align'
        focus_df = dfrm[
            (dfrm.sifts_range_tag.isin(('Deletion', 'Insertion & Deletion')))
            & (dfrm.repeated.eq(False))]
        focus = focus_df[['pdb_id', 'entity_id', 'UniProt']].to_numpy()
        focus_index = focus_df.index
        updated_pdb_range, updated_unp_range = list(), list()
        seqAligner = SeqPairwiseAlign()
        for record in focus:
            # pdbSeq = cls.get_seqres(record['pdb_id'], record['entity_id'])
            query, kwargs = cls.get_seqres(record[0], record[1])
            res = await neo4j_api.afetch(query, **kwargs)
            pdbSeq = ''.join(SEQ_DICT.get(r['residue_name'], 'X') for r in res)
            # unpSeq = cls.get_unp_seq(record["UniProt"])
            query, kwargs = cls.get_unp_seq(record[2])
            res = await neo4j_api.afetch(query, **kwargs)
            unpSeq = ''.join(r['aa'] for r in res)
            # ---
            res = seqAligner.makeAlignment(unpSeq, pdbSeq)
            updated_unp_range.append(res[0])
            updated_pdb_range.append(res[1])

        updated_range_df = pd.DataFrame(
            {new_unp_range: updated_unp_range, new_pdb_range: updated_pdb_range}, index=focus_index)
        dfrm = pd.merge(dfrm, updated_range_df, left_index=True,
                        right_index=True, how='left')
        dfrm[new_unp_range] = dfrm.apply(lambda x: x['unp_range'] if pd.isna(
            x[new_unp_range]) else x[new_unp_range], axis=1)
        dfrm[new_pdb_range] = dfrm.apply(lambda x: x['pdb_range'] if pd.isna(
            x[new_pdb_range]) else x[new_pdb_range], axis=1)
        return dfrm

    @classmethod
    @lru_cache()
    def get_unp_seq(cls, unp: str, session=None):
        '''
        TODO: Different behavior
        '''
        query = '''
            MATCH (unp:UniProt{ACCESSION: "%s"})-[:HAS_UNP_RESIDUE]-(unpRes:UNPResidue)
            RETURN unpRes.ONE_LETTER_CODE AS aa ORDER BY toInteger(unpRes.ID)
        ''' % unp
        if session is None:
            if cls.session is not None:
                session = cls.session
        try:
            return ''.join(r['aa'] for r in session.run(query))
        except AttributeError:
            return query, {}
    
    @classmethod
    def get_unp_seq_len(cls, unp: str, session=None):
        query = '''
            MATCH (unp:UniProt{ACCESSION: "%s"})-[:HAS_UNP_RESIDUE]-(unpRes:UNPResidue)
            RETURN COUNT(unpRes) as unp_len
        ''' % unp
        if session is None:
            if cls.session is not None:
                session = cls.session
        try:
            return session.run(query)
        except AttributeError:
            return query, {}

    @classmethod
    def res_conflict_then(cls, res):
        def get_mutaRes(dfrm: pd.DataFrame):
            sites = defaultdict(list)
            def storeSites(lyst, sitelyst):
                sitelyst.extend([i.split('|')[1] for i in lyst if '|' in i])# !!!
            dfrm.apply(lambda x: storeSites(x['resCon.DETAILS'], sites[(x['pdb_id'], x['entity_id'], x['resCon.ID'])]), axis=1)
            return sites

        def yield_dfrm_of_mutaRes(sites: Dict):
            for (pdb, entity, tag), value in sites.items():
                cur_df = pd.DataFrame(set(value), columns=['residue_number'])
                cur_df['pdb_id'] = pdb
                cur_df['entity_id'] = entity
                cur_df['tag'] = tag
                yield cur_df
        
        result = cls.to_data_frame(res)
        if len(result) == 0:
            return None
        try:
            sites = get_mutaRes(result)
        except IndexError:
            logging.error(result.to_string())
            return None
        return pd.concat(yield_dfrm_of_mutaRes(sites), sort=False, ignore_index=True)

    @classmethod
    def summary_res_conflict(cls, lyst, session=None):
        muta_query = '''
            MATCH (entry:Entry)-[:HAS_ENTITY]->(entity:Entity{POLYMER_TYPE:'P'})-[:HAS_RESIDUE_CONFLICT]->(resCon:ResidueConflict)
            WHERE entry.ID in $lyst
            RETURN entry.ID as pdb_id, entity.ID as entity_id, resCon.DETAILS, resCon.ID
        '''
        if session is None:
            if cls.session is not None:
                session = cls.session
        try:
            res = session.run(muta_query, lyst=list(lyst))
            return cls.res_conflict_then(res)
        except AttributeError:
            return muta_query, dict(lyst=list(lyst))

    @classmethod
    def deal_res_conflict(cls, res: pd.DataFrame):
        res.tag = res.tag.apply(lambda x: x.split('_')[0])
        group_cols = ['pdb_id', 'entity_id', 'tag']
        return pd.DataFrame(
            ((pdb_id, entity_id, tag, 
              json.dumps(to_interval(groupData.residue_number)).decode('utf-8')
              ) for (pdb_id, entity_id, tag), groupData in res.groupby(group_cols)),
            columns=group_cols+['conflict_range'])

    @staticmethod
    def yieldPureHo(entity_chain_map: Dict, unp_entity_info: Dict):
        for unp, entities in unp_entity_info.items():
            for entity in entities:
                chains = entity_chain_map[entity]
                if len(chains) > 1:
                    yield unp, tuple(combinations(chains, 2))

    @staticmethod
    def pass_range_check(left, right):
        print(
            jaccard.similarity(lyst2range(left), lyst2range(right)),
            overlap.similarity(lyst2range(left), lyst2range(right)))
        return True

    @classmethod
    def yieldDetectHo(cls, entity_chain_map: Dict, unp_entity_info: Dict):
        for unp, entities in unp_entity_info.items():
            if len(entities) > 1:
                for entity_a, entity_b in combinations(entities, 2):
                    if cls.pass_range_check(entities[entity_a], entities[entity_b]):
                        yield unp, tuple(product(entity_chain_map[entity_a], entity_chain_map[entity_b]))
    
    @staticmethod
    def yieldHe(entity_chain_map: Dict, unp_entity_info: Dict):
        for unp_a, unp_b in combinations(unp_entity_info, 2):
            # Check Taxonomy
            if unp_a[1] != unp_b[1]:
                continue
            for entity_a, entity_b in product(unp_entity_info[unp_a], unp_entity_info[unp_b]):
                assert entity_a != entity_b
                yield (unp_a, unp_b), tuple(product(entity_chain_map[entity_a], entity_chain_map[entity_b]))


class SeqPairwiseAlign(object):
    def __init__(self):
        self.seqa = None
        self.seqb = None
        self.aligner = Align.PairwiseAligner()
        self.aligner.mode = 'global'
        self.aligner.open_gap_score = -10
        self.aligner.extend_gap_score = -0.5
        self.aligner.substitution_matrix = matlist.blosum62
        self.alignment_count = 0

    @lru_cache()
    def makeAlignment(self, seqa, seqb):
        if seqa is None or seqb is None:
            return np.nan, np.nan
        self.seqa = seqa
        self.seqb = seqb
        alignments = self.aligner.align(seqa, seqb)
        for alignment in alignments:
            result = self.getAlignmentSegment(alignment)
            self.alignment_count += 1
            return json.dumps(result[0]).decode('utf-8'), json.dumps(result[1]).decode('utf-8')

    @staticmethod
    def getAlignmentSegment(alignment):
        segments1 = []
        segments2 = []
        i1, i2 = alignment.path[0]
        for node in alignment.path[1:]:
            j1, j2 = node
            if j1 > i1 and j2 > i2:
                segment1 = (i1 + 1, j1)
                segment2 = (i2 + 1, j2)
                segments1.append(segment1)
                segments2.append(segment2)
            i1, i2 = j1, j2
        return segments1, segments2


@lru_cache()
def convert_index(lrange_str: str, rrange_str: str, site: int):
    lrange = json.loads(lrange_str)
    rrange = json.loads(rrange_str)
    try:
        for (lstart, lend), (rstart, rend) in zip(lrange, rrange):
            assert (lend - lstart) == (rend-rstart), "convert_index(): Invalid range"
            if (site >= rstart) and (site <= rend):
                return int(site + lstart - rstart)
            else:
                continue
        return -1
    except AssertionError as e:
        logging.error(e)
        raise e


class Neo4j_API(Abclog):
    
    output_kwargs = {'sep': '\t', 'index': False, 'header': False, 'mode': 'a'}
    filtered_sifts_path = None
    res2pdbpath = None
    neo4j_api = None
    sqlite_api = None
    unpmap_filter:Dict = None
    sifts_filter:Dict = None
    entry_filter:Dict = None
    entry_info_all = ('method', 'ligand', 'nucleotides')
    entry_info_funcs = dict(zip(
        entry_info_all,
        (Entry.summary_method,Entry.summary_ligand,Entry.summary_nucleotides)))
    entry_info_add = {
        'has_hybrid_nucleotides': False,
        'nucleotides_entity_type': '',
        'BOUND_LIGAND_COUNT': 0,
        'BOUND_MOL_COUNT': 0
        }

    @classmethod
    @unsync
    async def pipe_new_sifts(cls, lyst, id_type):
        if len(lyst) > 0:
            # Assume that the elements of the lyst are all valid and are covered by the remote db
            query, kwargs = SIFTS.summary_mapping(lyst, id_type)
            res = await cls.neo4j_api.afetch(query, **kwargs)
            sifts_df = await SIFTS.deal_mapping(res, cls.neo4j_api)
            if sifts_df is not None:
                sifts_df = related_dataframe(cls.sifts_filter, sifts_df)
                if len(sifts_df) > 0:
                    # TODO: ADD Residue Conflict Info
                    query, kwargs = SIFTS.summary_res_conflict(
                        sifts_df.pdb_id.unique())
                    res = await cls.neo4j_api.afetch(query, **kwargs)
                    res = SIFTS.res_conflict_then(res)
                    if res is not None:
                        conflict_df = SIFTS.deal_res_conflict(res)
                        sifts_df = sifts_df.merge(conflict_df.rename(
                            columns={'tag': 'UniProt'}), how='left')
                    else:
                        sifts_df['conflict_range'] = None
                    # TODO: Export New Data [Sync]
                    values = sifts_df.to_dict('records')
                    await cls.sqlite_api.async_insert(cls.sqlite_api.SIFTS_Info, values)
                    return sifts_df
                else:
                    return None
            else:
                return None
        else:
            return None

    @classmethod
    @unsync
    async def pdb2unp(cls, lyst: Union[Iterable, Unfuture]):
        if not isinstance(lyst, Iterable):
            lyst = await lyst
        # Can be optimized
        sifts_df = await cls.pipe_new_sifts(lyst, 'pdb')
        return sifts_df

    @classmethod
    @unsync
    async def amapres2unp(cls, site_df: pd.DataFrame, unp:str, pdb_id: str, entity_id: str, chain_id: str, pdb_range: str, unp_range: str):
        sqlite_api = cls.sqlite_api
        sites = [convert_index(unp_range, pdb_range, x) for x in site_df.Pos]
        site_df['unp_index'] = sites
        res = await sqlite_api.PDBRes_Info.objects.filter(
            pdb_id=pdb_id, entity_id=entity_id,
            chain_id=chain_id, author_residue_number__in=site_df.Pos.unique()).all()
        pass

    @classmethod
    @unsync
    async def process_map2unp(cls, df, respath, siftspath, observed_only):
        df.pdb_id = df.pdb_id.str.lower()
        if 'entity_id' in df.columns:
            df.entity_id = df.entity_id.astype(str)
        query, kwargs = Entry.summary_eec(df.pdb_id.unique())
        res = await cls.neo4j_api.afetch(query, **kwargs)
        df = df.merge(Entry.to_data_frame(res), how="left")
        pdb_entity_struct_asym_id = df.pdb_id + '_' + df.entity_id + '_' + df.struct_asym_id
        query, kwargs = SIFTS.pdbchain2unpres(pdb_entity_struct_asym_id.unique(), observed_only)
        res = await cls.neo4j_api.afetch(query, **kwargs)
        res = SIFTS.to_data_frame(res)
        if len(res) > 0:
            res[['UniProt', 'unp_index']] = res.unp_index.str.split('_', expand=True)
            res.UNIQID = res.apply(lambda x: '_'.join(x['UNIQID'].split('_')[0:2]+[x['chain_id']]), axis=1)
            res.drop(columns=['chain_id'], inplace=True)
            res.UNIQID = res.UNIQID + '_' + res.UniProt
            res['sifts_check'] = 0
            sifts_df = await cls.pdb2unp(df.pdb_id.unique())
            sifts_df['UNIQID'] = sifts_df.pdb_id + '_' + sifts_df.entity_id.astype(str) + '_' + sifts_df.chain_id + '_' + sifts_df.UniProt
            focus_index = res[res.UNIQID.isin(set(res.UNIQID) & set(sifts_df.UNIQID))].index
            res.loc[focus_index, 'sifts_check'] = 1
            await pipe_out(res, respath)
            await pipe_out(sifts_df, siftspath)

    @classmethod
    @unsync
    async def pipe_new_entry_info(cls, pdbs):
        entry_info_func_res: Dict = {}
        for key, func in cls.entry_info_funcs.items():
            query, kwargs = func(pdbs)
            res = await cls.neo4j_api.afetch(query, **kwargs)
            entry_info_func_res[key] = Entry.to_data_frame(res)
        res = Entry.deal_nucleotides(entry_info_func_res.get('nucleotides', None))
        if res is not None:
            entry_info_func_res['nucleotides'] = res
        merge = partial(pd.merge, on=['pdb_id'], how='outer')
        pdb_entry_df = reduce(merge, (df for df in entry_info_func_res.values() if len(df) > 0))
        pdb_entry_df.rename(columns={'index': 'pdb_id'}, inplace=True)
        for col, default in cls.entry_info_add.items():
            if col not in pdb_entry_df:
                pdb_entry_df[col] = default
            else:
                pdb_entry_df[col].fillna(default, inplace=True)
        pdb_entry_df.BOUND_LIGAND_COUNT = pdb_entry_df.BOUND_LIGAND_COUNT.apply(int)
        pdb_entry_df.BOUND_MOL_COUNT = pdb_entry_df.BOUND_MOL_COUNT.apply(int)
        return pdb_entry_df

    @classmethod
    @unsync
    async def unp2pdb(cls, lyst: Union[Iterable, Unfuture]):
        '''
        Map from unp to pdb [DB]
        
        Principle
        ---------
        * Keep Consistency
        * Only Save What We Want [Filtered Data]
        '''
        if not isinstance(lyst, Iterable):
            lyst = await lyst  # .result()
        sqlite_api = cls.sqlite_api
        
        # TODO: Load Existing Data
        exists = await sqlite_api.SIFTS_Info.objects.filter(UniProt__in=lyst).all()
        if exists:
            e_sifts_df = pd.DataFrame(exists)
            lyst = list(set(lyst) - set(e_sifts_df.UniProt))
            # Get pdb lyst that already been saved in local db
            # e_pdbs = e_sifts_df.pdb_id.unique()
        else:
            e_sifts_df = None
        # TODO: Fetch New Data
        sifts_df = await cls.pipe_new_sifts(lyst, 'unp')
        sifts = [df for df in (sifts_df, e_sifts_df) if df is not None]
        if sifts:
            sifts_df = pd.concat(sifts, sort=False, ignore_index=True)
        else:
            return None
        # Following the principle of KEEP CONSISTENCY, we does not need the code below
        # e_sifts_df = related_dataframe(cls.sifts_filter, e_sifts_df)
        # Renew the pdb lyst. Now the elements in the lyst are all new to the local db
        # pdbs = list(set(pdbs)-set(e_pdbs))
        # TODO: Fetch Existing Data
        pdbs = sifts_df.pdb_id.unique()
        exists = await sqlite_api.Entry_Info.objects.filter(pdb_id__in=pdbs).all()
        if exists:
            e_pdb_entry_df = pd.DataFrame(exists)
            pdbs = list(set(pdbs) - set(e_pdb_entry_df.pdb_id))
        else:
            e_pdb_entry_df = None
        # TODO: Fetch New Data
        if len(pdbs) > 0:
            pdb_entry_df = await cls.pipe_new_entry_info(pdbs)
            pdb_entry_df = related_dataframe(cls.entry_filter, pdb_entry_df)
            # TODO: Export New Data [Sync]
            values = pdb_entry_df.to_dict('records')
            if values:
                await sqlite_api.async_insert(
                    sqlite_api.Entry_Info, values)
                # pdbs = pdb_entry_df.pdb_id.unique()
            else:
                pdb_entry_df = None
                # pdbs = []
            # del values
        else:
            pdb_entry_df = None
        
        pdb_entries = [df for df in (pdb_entry_df, e_pdb_entry_df) if df is not None]
        if pdb_entries:
            pdb_entry_df = pd.concat(pdb_entries, sort=False, ignore_index=True)
        else:
            return None
        
        # TODO: Load Existing Data
        pdbs = pdb_entry_df.pdb_id.unique()
        exists = await sqlite_api.SEQRES_Info.objects.filter(pdb_id__in=pdbs).all()
        if exists:
            e_seq_res_df = pd.DataFrame(exists)
            pdbs = list(set(pdbs) - set(e_seq_res_df.pdb_id))
        else:
            e_seq_res_df = None
        # TODO: Fetch New Data
        if len(pdbs) > 0:
            query, kwargs = Entry.summary_seq(pdbs)
            res = await cls.neo4j_api.afetch(query, **kwargs)
            seq_res_df = SIFTS.deal_seq_index(SIFTS.to_data_frame(res))
            # WARNING: NEED TO BE FIXED, SHOULD BE MORE FLEXIBLE
            del_pdbs = seq_res_df[
                seq_res_df.AVG_OBS_OBS_RATIO.le(0.25) & seq_res_df.SEQRES_COUNT.ge(50)
            ].pdb_id.unique()
            seq_res_df = seq_res_df[~seq_res_df.pdb_id.isin(del_pdbs)]
            values = seq_res_df.to_dict('records')
            # TODO: Export New Data [Sync]
            if values:
                await sqlite_api.async_insert(sqlite_api.SEQRES_Info, values)
            else:
                seq_res_df = None
        else:
            seq_res_df = None
        
        seq_ress = [df for df in (seq_res_df, e_seq_res_df) if df is not None]
        if seq_ress:
            seq_res_df = pd.concat(seq_ress, sort=False, ignore_index=True)
        else:
            return None
        
        return sifts_df[sifts_df.pdb_id.isin(seq_res_df.pdb_id.unique())]

    @classmethod
    @unsync
    async def amapres2pdb(cls, site_df: pd.DataFrame, unp:str, pdb_id: str, entity_id: str, chain_id: str, pdb_range: str, unp_range: str):
        sqlite_api = cls.sqlite_api
        sites = [convert_index(pdb_range, unp_range, x) for x in site_df.Pos]
        site_df['residue_number'] = sites
        sites = list(set(sites) | set({1}))
        res = await sqlite_api.PDBRes_Info.objects.filter(
            pdb_id=pdb_id, entity_id=entity_id,
            chain_id=chain_id, residue_number__in=sites).all()
        if not res:
            query, kwargs = Entry.get_residues(
                pdb_id, entity_id,
                chain_id)  # , sites
            res = await cls.neo4j_api.afetch(query, **kwargs)
            res = Entry.to_data_frame(res)
            if len(res) > 0:
                res.author_insertion_code.fillna('', inplace=True)
                await sqlite_api.async_insert(
                    sqlite_api.PDBRes_Info, res.to_dict('records'))
            else:
                return None
        else:
            res = pd.DataFrame(res)
        merge_df = pd.merge(site_df, res)
        merge_df['UniProt'] = unp
        return merge_df

    @classmethod
    @unsync
    async def mapres2pdb(cls, unp_map_df: pd.DataFrame, sifts_df: pd.DataFrame) -> Union[Unfuture, str, Path]:
        if unp_map_df is None or sifts_df is None:
            return
        unp_map_df.sort_values(by='UniProt', inplace=True)
        sifts_df.sort_values(by='UniProt', inplace=True)
        sqlite_api = cls.sqlite_api
        focus_cols = ['pdb_id', 'entity_id', 'chain_id',
                      'new_pdb_range', 'new_unp_range']
        sifts_nda = sifts_df[focus_cols].to_numpy()
        yourlist = unp_map_df.yourlist.to_numpy()
        unp_map_i_dict = slice_series(unp_map_df.UniProt.to_numpy())
        sifts_dict = slice_series(sifts_df.UniProt.to_numpy())        
        try:
            res_lyst = []
            for unp, (start, end) in sifts_dict.items():
                focus_sifts_nda = sifts_nda[start:end]
                start, end = unp_map_i_dict[unp]
                from_lyst = np.unique(yourlist[start:end])
                site_df = await sqlite_api.Site_Info.objects.filter(from_id__in=from_lyst).all()
                site_df = pd.DataFrame(site_df)
                for record in focus_sifts_nda:
                    res = await cls.amapres2pdb(site_df, unp, *record)
                    if res is not None and len(res) > 0:
                        res_lyst.append(res)
            if res_lyst:
                res = pd.concat(res_lyst, sort=False, ignore_index=True)
                cols = sorted(res.columns)
                res = res[cols]
                if cls.res2pdbpath.exists():
                    headers = None
                else:
                    headers = cols
                async with aiofiles.open(cls.res2pdbpath, 'a') as fileOb:
                    dataset = Dataset(headers=headers)
                    dataset.extend(res.to_records(index=False))
                    await fileOb.write(dataset.export('tsv'))
                return 1
            return 0
        except Exception as e:
            logging.error("res_map exception: "+str(e))
            traceback.print_exc()
            raise e

    @classmethod
    @unsync
    async def process_unp2pdb(cls, indata: Union[str, Path, pd.DataFrame, Unfuture], sep='\t') -> Union[Unfuture, Tuple]:
        '''
        Default chain process
        Default file seperator: tab
        '''
        # TODO: Add Converters
        if not isinstance(indata, (str, Path, pd.DataFrame)):
            indata = await indata  # .result()
        if isinstance(indata, pd.DataFrame):
            unp_map_df = indata
        elif indata is None:
            return None, None
        else:
            unp_map_df = pd.read_csv(indata, sep=sep)
        unp_map_df = related_dataframe(cls.unpmap_filter, unp_map_df)
        related_unp = unp_map_df.UniProt.unique()
        sifts_df = await cls.unp2pdb(related_unp)
        if sifts_df is not None:
            unp_map_df = unp_map_df[unp_map_df.UniProt.isin(sifts_df.UniProt.to_numpy())].reset_index(drop=True)
            sifts_df = sifts_df.reset_index(drop=True)
            if cls.filtered_sifts_path is not None:
                cols = sorted(sifts_df.columns)
                sifts_df = sifts_df[cols]
                if cls.filtered_sifts_path.exists():
                    headers = None
                else:
                    headers = cols
                async with aiofiles.open(cls.filtered_sifts_path, 'a') as fileOb:
                    dataset = Dataset(headers=headers)
                    dataset.extend(sifts_df.to_records(index=False))
                    await fileOb.write(dataset.export('tsv'))
                    # await fileOb.flush()
            return unp_map_df, sifts_df
        else:
            return None, None

    @classmethod
    @unsync
    async def process_mapres2pdb(cls, dfs: Union[Tuple[pd.DataFrame], Unfuture]) -> Union[Unfuture, str, Path]:
        if isinstance(dfs, Tuple):
            unp_map_df, sifts_df = dfs
        else:
            unp_map_df, sifts_df = await dfs  # .result()
        return cls.mapres2pdb(unp_map_df, sifts_df)

    @classmethod
    @unsync
    async def pipe_api_info(cls, func, **kwargs):
        query, newkwargs = func(**kwargs)
        res = await cls.neo4j_api.afetch(query, **newkwargs)
        return Entry.to_data_frame(res)

    @classmethod
    @unsync
    async def load_interact_chain_info(cls, i3d, pdbs):
        pre_cols = ['from_pisa_bond', 'from_arp_contact', 'from_i3d_chain_pair']
        pisa_df = await cls.pipe_api_info(Entry.summary_chain_interaction, pdbs=pdbs, source='pisa_bond')
        arp_df = await cls.pipe_api_info(Entry.summary_chain_interaction, pdbs=pdbs, source='arp_contact')
        cols = ['chain_id_1', 'chain_id_2']
        dfs = [df for df in (sort_sub_cols(pisa_df, cols), sort_sub_cols(arp_df, cols), i3d) if len(df) > 0]
        if len(dfs) > 0:
            merge = partial(pd.merge, how='outer')
            res = reduce(merge, dfs)
            for col in pre_cols:
                if col not in res.columns:
                    res[col] = False
                else:
                    res[col] = res[col].fillna(False)
            return res[['pdb_id']+cols+pre_cols]
        else:
            return None

    @staticmethod
    def pipe_sifts2interact(sifts_df):
        sifts_df['Entry'] = sifts_df.UniProt.str.split('-').str[0]
        # sifts_df = sifts_df[['Entry', 'UniProt', 'pdb_id', 'chain_id']]
        cols = [i for i in sifts_df.columns if i != 'pdb_id']
        sifts_df1 = sifts_df.rename(columns=dict(zip(cols, (i+'_1' for i in cols))))
        sifts_df2 = sifts_df.rename(columns=dict(zip(cols, (i+'_2' for i in cols))))
        sifts_df1 = sifts_df1.rename(columns={"Entry_1": "PROT1"})
        sifts_df2 = sifts_df2.rename(columns={"Entry_2": "PROT2"})
        sifts_df12 = pd.merge(sifts_df1, sifts_df2)
        return sifts_df12[sifts_df12.chain_id_1 != sifts_df12.chain_id_2].reset_index(drop=True)

    @classmethod
    @unsync
    async def pipe_interaction(cls, pdbs, i3d):
        int_df = await cls.load_interact_chain_info(i3d.chain_pairs_p(pdbs), pdbs)
        sifts_df = await cls.pipe_new_sifts(pdbs, 'pdb') # ! FILTERED & Can be optimized
        sifts_df = cls.pipe_sifts2interact(sifts_df)
        return int_df.merge(sifts_df, how='left').merge(i3d.unp_chain_pairs_p(pdbs), how='left').merge(i3d.all_unp_pairs_12, how='left')





# @Created Date: 2020-01-18 10:53:07 am
# @Filename: webfetch.py
# @Email:  1730416009@stu.suda.edu.cn
# @Author: ZeFeng Zhu
# @Last Modified: 2020-02-09 08:50:23 pm
# @Copyright (c) 2020 MinghuiGroup, Soochow University
import os
from time import perf_counter
import asyncio
import aiohttp
import aioftp
import aiofiles
from unsync import unsync, Unfuture
from tenacity import retry, wait_random, stop_after_attempt, after_log, RetryError
import logging
from tqdm import tqdm
from typing import Iterable, Iterator, Union, Any, Optional, List, Dict, Coroutine, Callable
from furl import furl
from pdb_profiling.log import Abclog
import re
from pdb_profiling.utils import init_semaphore, unsync_wrap


class UnsyncFetch(Abclog):
    '''
    Fetch files through API in unsync/async way

    Note:

    * Implement both GET and POST method
    * Load all the available data(stream) in memory as soon as it is received
    * Since the methods in this class would not load the entire file in memory,
      the response data could be a large file
    * Parameters of `tenacity.retry` is built-in
    
    following packages provide me with a lot of insights and inspiration

    Reference 
    ---------

        `aiohttp`

        * https://github.com/aio-libs/aiohttp
        * https://docs.aiohttp.org/en/stable/client_quickstart.html
        * https://docs.aiohttp.org/en/stable/streams.html

        `unsync`

        * https://github.com/alex-sherman/unsync

        `tenacity`

        * https://github.com/jd/tenacity
        * https://tenacity.readthedocs.io/en/latest/
    '''

    use_existing: bool = False

    @classmethod
    @retry(wait=wait_random(max=20), stop=stop_after_attempt(5))
    async def http_download(cls, method: str, info: Dict, path: str):
        if cls.use_existing is True and os.path.exists(path) and (os.stat(path).st_size > 0):
            return path
        cls.logger.debug(f"Start to download file: {info['url']}")
        async with aiohttp.ClientSession() as session:
            async_func = getattr(session, method)
            async with async_func(**info) as resp:
                if resp.status == 200:
                    async with aiofiles.open(path, 'wb') as fileOb:
                        # Asynchronous iterator implementation of readany()
                        async for chunk in resp.content.iter_any():
                            await fileOb.write(chunk)
                    cls.logger.debug(f"File has been saved in: {path}")
                    return path
                elif resp.status in (403, 404, 405):
                    cls.logger.debug(f"403|404|405 for: {info}")
                    return None
                else:
                    mes = "code={resp.status}, message={resp.reason}, headers={resp.headers}".format(resp=resp)
                    cls.logger.error(mes)
                    cls.logger.error(info)
                    raise Exception(mes)

    @classmethod
    @retry(wait=wait_random(max=20), stop=stop_after_attempt(5))
    async def ftp_download(cls, method: str, info: Dict, path: str):
        url = furl(info['url'])
        fileName = url.path.segments[-1]
        filePath = os.path.join(path, fileName)
        if cls.use_existing is True and os.path.exists(filePath) and (os.stat(filePath).st_size > 0):
            return filePath
        cls.logger.debug(f"Start to download file: {url}")  # info
        async with aioftp.ClientSession(url.host) as session:
            await session.change_directory('/'.join(url.path.segments[:-1]))
            await session.download(fileName, path) # , write_info=True
        cls.logger.debug(f"File has been saved in: {filePath}")
        return filePath

    @classmethod
    def download_func_dispatch(cls, method: str):
        method = method.lower()
        if method in ('get', 'post'):
            return cls.http_download
        elif method == 'ftp':
            return cls.ftp_download
        else:
            raise ValueError(f'Invalid method: {method}, valid method should be get, post or ftp')

    @classmethod
    @unsync
    async def save_file(cls, path: str, data: bytes):
        '''Deprecated'''
        async with aiofiles.open(path, 'wb') as fileOb:
            await fileOb.write(data)

    @classmethod
    @unsync
    async def fetch_file(cls, semaphore, method: str, info: Dict, path: str, rate: float):
        download_func = cls.download_func_dispatch(method)
        try:
            async with semaphore:
                res = await download_func(method, info, path)
                if res is not None:
                    await asyncio.sleep(rate)
                return res
        except RetryError:
            cls.logger.error(f"Retry failed for: {info}")

    @classmethod
    @unsync
    async def unsync_tasks(cls, tasks):
        return [await fob for fob in tqdm(asyncio.as_completed(tasks), total=len(tasks))]

    @classmethod
    def single_task(cls, task, semaphore, to_do_func: Optional[Callable] = None, rate: float = 1.5) -> Unfuture:
        method, info, path = task
        if cls.use_existing is True and os.path.exists(path) and os.path.isfile(path) and (os.stat(path).st_size > 0):
            task = unsync_wrap(path)
        else:
            task = cls.fetch_file(semaphore, method, info, path, rate)
        if to_do_func is not None:
            task = task.then(to_do_func)
        return task

    @classmethod
    def multi_tasks(cls, tasks: Union[Iterable, Iterator],
              to_do_func: Optional[Callable] = None,
              concur_req: int = 4, rate: float = 1.5,
              ret_res: bool = True,
              semaphore = None
              ):

        if semaphore is None:
            # asyncio.Semaphore(concur_req)
            semaphore = init_semaphore(concur_req).result()
        else:
            cls.logger.debug(f'{cls.multi_tasks.__qualname__}: pass {repr(semaphore)}')
        if to_do_func is None:
            tasks = [cls.fetch_file(semaphore, method, info, path, rate)
                     for method, info, path in tasks]
        else:
            tasks = [cls.fetch_file(semaphore, method, info, path, rate).then(
                to_do_func) for method, info, path in tasks]
        if ret_res:
            t0 = perf_counter()
            res = cls.unsync_tasks(tasks).result()
            elapsed = perf_counter() - t0
            cls.logger.info('{} chunks downloaded in {:.2f}s'.format(len(res), elapsed))
            return res
        else:
            return tasks
        

    @classmethod
    def main(cls, workdir: str, data: Union[Iterable, Iterator], concur_req: int = 4, rate: float = 1.5, logName: str = 'UnsyncFetch'):
        cls.set_logging_fileHandler(os.path.join(workdir, f'{logName}.log'), logName=cls.__name__)
        t0 = perf_counter()
        # res = asyncio.run(cls.multi_tasks(data, concur_req=concur_req, rate=rate))
        res = cls.multi_tasks(data, concur_req=concur_req, rate=rate).result()
        elapsed = perf_counter() - t0
        cls.logger.info(f'downloaded in {elapsed}s')
        return res



# @Created Date: 2020-10-13 04:38:16 pm
# @Filename: cif_gz_stream.py
# @Email:  1730416009@stu.suda.edu.cn
# @Author: ZeFeng Zhu
# @Last Modified: 2020-10-13 04:38:24 pm
# @Copyright (c) 2020 MinghuiGroup, Soochow University
from aiohttp import ClientSession
from aiofiles import open as aiofiles_open
from unsync import unsync
import zlib
from typing import Iterable
from pdb_profiling.utils import MMCIF2DictPlus

'''
>>> pdb_id, assembly_id = '2o2q', 2
>>> header = f'{pdb_id}-assembly-{assembly_id}'

>>> writer(
        reader(f'http://www.ebi.ac.uk/pdbe/static/entry/download/{header}.cif.gz'), 
        f'{header}-pdbe_chain_remapping.cif',
        b'data_%s\n#\nloop_\n' % bytes(header, 'utf-8'),
        b'_pdbe_chain_remapping'
        ).result()

>>> parser(
        reader(f'http://www.ebi.ac.uk/pdbe/static/entry/download/{header}.cif.gz'),
        ('data_%s\n' % header, '#\n', 'loop_\n'),
        b'_pdbe_chain_remapping'
        )
>>> 
'''


def iter_index(text, target, add):
    '''
    >>> text = b'sdgfsd\nsdgsdg\nfdsg\nd'
    >>> index = (None, *iter_index(text), None)
    >>> print(index)
    >>> tuple(text[start:end] for start,end in zip(index, index[1:]))
    '''
    text_len = len(text)
    start = -1
    while True:
        try:
            res = text.index(target, start+1) + add
            yield res
            start = res
            if start >= text_len-1:
                break
        except ValueError:
            break


async def reader(url):
    dec = zlib.decompressobj(32 + zlib.MAX_WBITS)
    remain_part = b''
    async with ClientSession() as session:
        async with session.get(url=url) as resp:
            if resp.status == 200:
                async for chunk in resp.content.iter_any():
                    rv = dec.decompress(chunk)
                    if rv:
                        # yield rv
                        index = (None, *iter_index(rv, b'\n', 1), None)
                        if len(index) == 2:
                            remain_part += rv
                            continue
                        if remain_part:
                            yield remain_part + rv[:index[1]]
                            remain_part = b''
                            for start, end in zip(index[1:-1], index[2:-1]):
                                yield rv[start:end]
                        else:
                            for start, end in zip(index[:-1], index[1:-1]):
                                yield rv[start:end]
                        if index[-2] != len(rv):
                            remain_part = rv[index[-2]:]
                if remain_part:
                    yield remain_part
            else:
                raise Exception(
                    "code={resp.status}, message={resp.reason}, headers={resp.headers}".format(resp=resp))


@unsync
async def writer(handle, path, header:bytes, start_key:bytes):
    start = False
    async with aiofiles_open(path, 'wb') as fileOb:
        if header:
            await fileOb.write(header)
        async for line in handle:
            if line.startswith(start_key):
                start = True
            elif start and line.startswith(b'#'):
                await fileOb.write(line)
                return path
            if start:
                await fileOb.write(line)


def parser(handle, header_lines: Iterable[bytes], start_key: bytes):
    @unsync
    async def a_handle(handle):
        start = False
        res = []
        async for line in handle:
            if line.startswith(start_key):
                start = True
            elif start and line.startswith(b'#'):
                res.append(line.decode('utf-8'))
                return res
            if start:
                res.append(line.decode('utf-8'))

    return MMCIF2DictPlus(tuple(header_lines)+tuple(a_handle(handle).result()))
